from asap3 import EMT2013
from asap3.EMT2013Parameters import sihb_PtY_parameters
from ase.build import  bulk 
from asap3.test.pytest_markers import ReportTest, serial

import pytest

def make_Pt():
    return bulk('Pt', 'fcc', orthorhombic=True).repeat((6,6,6))

def make_Y():
        return bulk('Y', 'hcp', orthorhombic=True).repeat((6,6,6))

@serial
def test_EMT2013elements():
    atoms = make_Pt()
    atoms.calc = EMT2013(sihb_PtY_parameters, True)
    e_Pt1 = atoms.get_potential_energy()
    print("Pt (1):", e_Pt1)

    atoms = make_Y()
    atoms.calc = EMT2013(sihb_PtY_parameters, True)
    e_Y1 = atoms.get_potential_energy()
    print("Y (1):", e_Y1)

    calc = EMT2013(sihb_PtY_parameters)

    atoms = make_Pt()
    atoms.calc = calc
    e_Pt2 = atoms.get_potential_energy()
    print("Pt (1):", e_Pt2)

    atoms = make_Y()
    atoms.calc = calc
    e_Y2 = atoms.get_potential_energy()
    print("Y (1):", e_Y2)

    ReportTest("Energy of first element (Pt)", e_Pt2, e_Pt1, 1e-9)
    ReportTest("Energy of second element (Y)", e_Y2, e_Y1, 1e-9)



from asap3 import *
from ase.cluster.cubic import FaceCenteredCubic
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3.analysis import CoordinationNumbers
from ase.parallel import world

from asap3.test.pytest_markers import ReportTest, parallel

import pytest

@pytest.mark.skipif(world.size != 2, reason='Requires 2 MPI tasks')
def test_delete(cpulayout):
    ismaster = world.rank == 0

    # We need an asymmetric nanoparticle
    surfaces = [(1, 0, 0), (1, 1, 1), (0, -1, 0), (-1, 0, 0)]
    layers = [9, 7, 7, 8]
    lc = 4.08000
    if ismaster:
        atoms = FaceCenteredCubic('Au', surfaces, layers, latticeconstant=lc)
        atoms.center(vacuum=3.0)
    else:
        atoms = None

    atoms = MakeParallelAtoms(atoms, cpulayout)
    MaxwellBoltzmannDistribution(atoms, temperature_K=600)
    atoms.calc = EMT()

    dyn = Langevin(atoms, 0.5*units.fs, temperature_K=600, friction=0.01)
    dyn.run(1)

    natoms = atoms.get_global_number_of_atoms()
    zap = [0, natoms//2, natoms-1]
    print(f"Atoms on {world.rank} (before):", len(atoms))
    atoms.delete_atoms_globally(global_indices=zap)
    print(f"Atoms on {world.rank} (after):", len(atoms))

    ReportTest("Number of atoms reduced by 3", atoms.get_global_number_of_atoms(), natoms - len(zap), 0)
    cn = CoordinationNumbers(atoms)
    ReportTest("Number of coordination numbers", len(cn), len(atoms), 0)

    dyn.run(15)
    print("Simulation 1 completed without crashing.")

    natoms = atoms.get_global_number_of_atoms()

    # Try with different info on different tasks.
    if world.size == 2:
        if world.rank == 0:
            zap = [0, natoms//2]
        else:
            zap = [natoms//2, natoms-1]
    else:
        if world.rank == 1:
            zap = (natoms//2, natoms-1)
        elif world.rank == 2:
            zap = np.array([natoms//2, 0])
        else:
            zap = []
    atoms.delete_atoms_globally(global_indices=zap)

    dyn.run(15)

    ReportTest("Number of atoms reduced by 3 (inhomogeneous case)", atoms.get_global_number_of_atoms(), natoms - 3, 0)
    print("Simulation 2 completed without crashing.")



"""An observer keeping track of running average positions of atoms.

"""

from asap3.Internal.Subject import Subject

class RunningAverage:
    """Calculate running average positions of atoms.
    
    Attach this observer to the atoms to make average positions available
    e.g. for CNA analysis.  The filter is called often (maybe every 5 time
    steps).  After it has been called interval times (or a multiple hereof)
    the average positions are calculated, and the attached analysis 
    functions are called.  numavg specifies over how many calls are used
    to calculate the average positions.  If numavg==interval, all calls
    are used to collect data, and the average is over all times since the
    last analysis.  If numavg<interval then only the last numavg calls
    before the analysis is used to collect data.  Currently, numavg>interval
    is not supported.
    
    During analysis, positions of the atoms are set to the average positions.
    After analysis, observers can be called e.g. to store the results.  
    Observers may be called before or after the positions are restored to
    their instantaneous values.
    
    Warning: As the atoms are changed before and after analysis, migration
    may be triggered in parallel simulations.

    Parameters:

    atoms: 
        The atoms being observed.  Temporary data is stored in the atoms.
    
    interval:
        How often analysis is done.
    
    numavg:
        How many frames should be used for the averaging.
    
    Important methods:

    update(): 
        Update the running average, and do analysis if needed.  This
        is the function that should be attached to the dynamics.

    __call__(): 
        Same as update()
    
    attach_analysis(callable, *args, **kwargs): 
        Attach an analysis function.  Arguments are passed on.
        
    attach_observer(callable, interval=1, *args, **kwargs):
        Call an external observer for example to store the result
        of the analysis.  The observer is called every interval'th
        time the analysis is perfomed, it is rare to set interval
        different from 1.  The observer is called after the 
        positions have been restored to their instantaneous
        values.
        
    attach_observer_avgpos(callable, interval=1, *args, **kwargs):
        As attach_observer(), except that the observer is called while
        the positions still have their average value.
    """
    
    def __init__(self, atoms, interval, numavg):
        self.atoms = atoms
        self.interval = interval
        self.numavg = numavg
        if numavg > interval:
            raise NotImplementedError("The case numavg>interval has not yet been implemented.")
        self.firstdata = self.interval - self.numavg
        self.analysers = Subject()
        self.observers = Subject()
        self.avgobservers = Subject()
        self.extra_data_buffer = []
        self.n = 0
        self.m = 0
        
    def attach_analysis(self, func, *args, **kwargs):
        "Attach an analysis method"
        self.analysers.attach(func, 1, *args, **kwargs)
        
    def attach_observer(self, func, interval=1, *args, **kwargs):
        "Attach an observer called after analysis"
        if self.extra_data_buffer and hasattr(func, 'set_extra_data'):
            for name, source, once in self.extra_data_buffer:
                func.set_extra_data(name, source, once)
        self.observers.attach(func, interval, *args, **kwargs)
        
    def attach_observer_avgpos(self, func, interval=1, *args, **kwargs):
        "Attach an observer called after analysis but while positions remain averaged."
        self.avgobservers.attach(func, interval, *args, **kwargs)
        
    def reset(self):
        "Reset internal variables after an analysis"
        self.n = 0
        self.m = 0
        if self.atoms.has('avgpositions'):
            self.atoms.set_array('avgpositions', None)
        if self.atoms.has('sumpositions'):
            # This should not normally happen.
            self.atoms.set_array('sumpositions', None)
        
    def update(self):
        self.n += 1
        if self.n <= self.firstdata:
            return  # Do nothing
        # Take data, and include in average
        if self.atoms.has('sumpositions'):
            sp = self.atoms.get_array('sumpositions') + self.atoms.get_positions()
        else:
            sp = self.atoms.get_positions()
        self.atoms.set_array('sumpositions', sp)
        self.m += 1
        if self.n == self.interval:
            assert self.m == self.numavg
            self.atoms.set_array('avgpositions', sp * (1.0/self.m))
            self.atoms.set_array('sumpositions', None)
            # Now save the instantanous positions and set the positions to
            # their average
            self.atoms.set_array('instant_positions', self.atoms.get_positions())
            self.atoms.set_positions(self.atoms.get_array('avgpositions'))
            # Call analysis and then observers (avg pos)
            self.analysers.call_observers()
            self.avgobservers.call_observers()
            # Reset positions to instantaneous value
            self.atoms.set_positions(self.atoms.get_array('instant_positions'))
            self.atoms.set_array('instant_positions', None)
            self.observers.call_observers()
            self.reset()
            
    __call__ = update
        
    def set_extra_data(self, name, source=None, once=False):
        """Forward to observers that extra data should be stored.

        This function is intended to make sure that if a BundleTrajectory
        is attached to an NPT dynamics through a RunningAverage object,
        then information about the dynamics is still saved in the 
        BundleTrajectory.

        It works by forwaring the call to set_extra_data to any observers
        having a set_extra_data method - whether they are attached before
        or after calling this function.
        """
        self.extra_data_buffer.append((name, source, once))
        for obs in self.observers.observers:
            f = obs[0]
            if hasattr(f, "set_extra_data"):
                f.set_extra_data(name, source, once)



# A Script using the downhill simplex algorithm to optimize the seven parameters in the EMT calculator

# Written by: Rasmus E. Christiansen

import Optimize as Optim
from ErrorFunction import ErrorFunction as EF
import numpy


# Previously fittet Values for parameters:
parameters = {
#          E0     s0    V0     eta2    kappa   lambda  n0      
#          eV     bohr  eV     bohr^-1 bohr^-1 bohr^-1 bohr^-3 
    'H':  (-2.21, 0.71, 2.132, 1.652,  2.790,  1.892,  0.00547),
    'Al': (-3.28, 3.00, 1.493, 1.240,  2.000,  1.169,  0.00700),
    'Cu': (-3.51, 2.67, 2.476, 1.652,  2.740,  1.906,  0.00910),
    'Ag': (-2.96, 3.01, 2.132, 1.652,  2.790,  1.892,  0.00547),
    'Au': (-3.80, 3.00, 2.321, 1.674,  2.873,  2.182,  0.00703),
    'Ni': (-4.44, 2.60, 3.673, 1.669,  2.757,  1.948,  0.01030),
    'Pd': (-3.90, 2.87, 2.773, 1.818,  3.107,  2.155,  0.00688),
    'Pt': (-5.85, 2.90, 4.067, 1.812,  3.145,  2.192,  0.00802),
    'C':  (-1.97, 1.18, 0.132, 3.652,  5.790,  2.892,  0.01322),
    'N':  (-4.97, 1.18, 0.132, 2.652,  3.790,  2.892,  0.01222),
    'O':  (-2.97, 1.25, 2.132, 3.652,  5.790,  4.892,  0.00850)}



# The parameters which are to be optimized are specified. This is done by initializing the dictionary 
# VariableParameters indexed by the elements whose parameters are being optimized including len()=8 arrays 
# of the form [0.,0.,1.,1.,0.,1.,0.], where 0 means the parameter is to be kept constant while 1 means it 
# is to be optimized. The Initial values of the parameters to be optimized are then set in the list ParamInit.
VariableParameters = {'Cu': numpy.array([1.,1.,1.,1.,1.,1.,1.]),
                      'Au': numpy.array([1.,1.,1.,1.,1.,1.,1.])}
Elements = VariableParameters.keys()
# The alloys desired are given in ElementsAllyos
#ElementsAlloys = ['AuCu3']
ElementsAlloys = []


# Experimental values picked from tabel
ParamInit = []
for i in Elements:
    for j in range(7):
        if VariableParameters[i][j] == 1:
            ParamInit.append(parameters[i][j])

# The ErrorFunction object is created using the chosen elements and variable parameters
ErrorFunc = EF(Elements,ElementsAlloys,VariableParameters)

# The minimization simplex algorithm is run using the ErrorFunction.
print(Optim.fmin(ErrorFunc.ErFu,ParamInit,xtol=0.001,ftol=0.001))






import numpy as np
from asap3.test.npt_module import main, bulkmodulus, MDLogger
try:
    from asap3.md.nose_hoover_chain import IsotropicMTKNPT
except ImportError:
    IsotropicMTKNPT = None
from asap3.md.nptberendsen import NPTBerendsen

from asap3.test.pytest_markers import serial, parallel

import pytest

@pytest.mark.slow
@serial
@pytest.mark.skipif(IsotropicMTKNPT is None, reason='ASE is too old')
def test_npt_isotropic_mtk():
    def mktdynmaker(atoms, T0, p0, dt, *, taut, taup, rng, logint):
        dyn = IsotropicMTKNPT(atoms, timestep=dt, temperature_K=T0, pressure_au=p0, 
                            tdamp=taut, pdamp=taup)
        log = MDLogger(dyn, atoms, '-', peratom=True, stress=True)
        dyn.attach(log, interval=logint)
        return dyn

    def berdynmaker(atoms, T0, p0, dt, *, taut, taup, rng, logint):
        dyn = NPTBerendsen(atoms, timestep=dt, temperature_K=T0, pressure_au=p0, 
                        taut=taut, taup=taup, compressibility_au=1 / bulkmodulus)
        log = MDLogger(dyn, atoms, '-', peratom=True, stress=True)
        dyn.attach(log, interval=logint)
        return dyn

    rng = np.random.default_rng(3141592)
    main(mktdynmaker, rng, initdyn=berdynmaker, failfluct=False)


@pytest.mark.slow
@parallel
@pytest.mark.skipif(IsotropicMTKNPT is None, reason='ASE is too old')
def test_par_npt_isotropic_mtk():
    def mktdynmaker(atoms, T0, p0, dt, *, taut, taup, rng, logint):
        dyn = IsotropicMTKNPT(atoms, timestep=dt, temperature_K=T0, pressure_au=p0, 
                            tdamp=taut, pdamp=taup)
        log = MDLogger(dyn, atoms, '-', peratom=True, stress=True)
        dyn.attach(log, interval=logint)
        return dyn

    def berdynmaker(atoms, T0, p0, dt, *, taut, taup, rng, logint):
        dyn = NPTBerendsen(atoms, timestep=dt, temperature_K=T0, pressure_au=p0, 
                        taut=taut, taup=taup, compressibility_au=1 / bulkmodulus)
        log = MDLogger(dyn, atoms, '-', peratom=True, stress=True)
        dyn.attach(log, interval=logint)
        return dyn

    rng = np.random.default_rng(314159265)
    main(mktdynmaker, rng, initdyn=berdynmaker, failfluct=True, parallel=True)




print("Check that it is possible to load ASE before Asap without messing up ase.parallel.world.")

from  ase.parallel import world as ase_world
from  asap3.mpi import world as asap_world
assert(ase_world.size == asap_world.size)



import numpy as np
from asap3.test.nvt_module import main
try:
    from asap3.md.bussi import Bussi
except ImportError:
    Bussi = None

from asap3.test.pytest_markers import serial, parallel
from ase.parallel import world

import pytest

@pytest.mark.slow
@serial
@pytest.mark.skipif(Bussi is None, reason='ASE is too old')
def test_nvt_bussi():
    def busdynmaker(atoms, T0, dt, tau, rng, logint):
        # tau is the energy relaxation time.  The velocity relaxation time
        # should be the double.
        return Bussi(atoms, dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint, rng=rng)

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()

    main(busdynmaker, rng, sloppytime=True)
    

# @pytest.mark.slow     # Not maked as slow so a single NVT test runs normally
@parallel
@pytest.mark.skipif(Bussi is None, reason='ASE is too old')
def test_par_nvt_bussi():
    def busdynmaker(atoms, T0, dt, tau, rng, logint):
        # tau is the energy relaxation time.  The velocity relaxation time
        # should be the double.
        dyn = Bussi(atoms, dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint, rng=rng)
        assert dyn.comm.size == world.size
        return dyn

    seed = np.random.SeedSequence(2718281828459045)
    seed = seed.spawn(world.size)[world.rank]
    rng = np.random.default_rng(seed)   # Use a fixed seed for reproducability

    main(busdynmaker, rng, parallel=True, sloppytime=True, failfluct=True)
    # Note: Fluctuation test always fail in parallel, as the boundaries
    # are free in order to test the migration better.  But that changes the
    # heat capacity, making the test fail.



#PBS -N longMelting
#PBS -q long
#PBS -l nodes=1:ppn=4:opteron4
#PBS -m ae

from asap3 import *
from asap3.md.langevin import Langevin
from asap3.md.nptberendsen import Inhomogeneous_NPTBerendsen
from asap3.md import MDLogger
from asap3.testtools import ReportTest
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
from asap3.mpi import world
import random
import time
import shutil
import os

quickie = False  # Runs quickly, but fails.
cleanup = True  # Remove output directory when done.

if world.size == 2:
    cpulayout1 = (1,2,1)
    cpulayout2 = (1,1,2)
elif world.size == 4:
    cpulayout1 = (1,2,2)
    cpulayout2 = (1,1,4)
elif world.size == 8:
    cpulayout1 = (2,1,4)
    cpulayout2 = (1,2,4)
elif world.size == 16:
    cpulayout1 = (2,2,4)
    cpulayout2 = (1,2,8)
else:
    raise RuntimeError("Not prepared to run on %s CPUs.  Supported: 2, 4, 8, 16." 
                       % (world.size,))

size = (10,10,50)
if quickie:
    steps1 = 1000
    steps2 = 1000
    takedata = 500
else:
    steps1 = 30000  # Phase 1
    steps2 = 1000000 # Phase 2
    takedata = 250000

friction = 0.0005

master = world.rank == 0


class Thermometer:
    def __init__(self, atoms, filename, blocks=50):
        self.atoms = atoms
        self.output = open(filename, "w")
        self.blocks = blocks

    def update(self):
        zmax = self.atoms.get_cell()[2,2]
        z = self.atoms.get_positions()[:,2]
        group = np.round(z/zmax * self.blocks).astype(int)
        group = np.clip(group, 0, self.blocks-1)
        p = self.atoms.get_momenta()
        ekin = (p*p).sum(axis=1) / (2 * self.atoms.get_masses())
        ekinsum = np.zeros(self.blocks)
        blocksum = np.zeros(self.blocks, int)
        for i in range(self.blocks):
            thisblock = np.equal(group, i)
            ekinsum[i] = ekin[thisblock].sum()
            blocksum[i] = thisblock.sum()
        world.sum(ekinsum)
        world.sum(blocksum)
        if master:
            np.clip(blocksum, 1, blocksum.max(), out=blocksum)
            T = ekinsum / ( 1.5 * units.kB * blocksum)
            for i in range(self.blocks):
                self.output.write("%i %.1f\n" % (i, T[i],))
            self.output.write("&\n")
            self.output.flush()
            
class AveragingThermometer:
    def __init__(self, atoms):
        self.atoms = atoms
        self.Tsum = 0
        self.n = 0
        
    def update(self):
        self.Tsum += self.atoms.get_temperature()
        self.n += 1
        
    def get_temp(self):
        return self.Tsum / self.n

def main(element, T_up, T_low, T_expect, bulkmodulus, makepot):
    def set_temp_fric(dyn, atoms):
        z = atoms.get_positions()[:,2]
        zmax = atoms.get_cell()[2,2]
        zmin = 0
        T = T_low + (T_up - T_low) * (z - zmin) / (zmax - zmin)
        dyn.set_temperature(temperature_K=T)

    rnd = world.sum(random.randint(0,2**32)) # Insures the same number on all nodes.
    prefix = f"melt-{element}-{hex(rnd)[2:]}"
    if master:
        print("Using output directory", prefix)
        os.mkdir(prefix)
    else:
        while not os.path.exists(prefix):
            time.sleep(1)
    if master:
        atoms = FaceCenteredCubic(symbol=element, size=size, pbc=(True, True, False))
        atoms.center(vacuum=10.0, axis=2)
    else:
        atoms = None
    atoms = MakeParallelAtoms(atoms, cpulayout1)
    print(world.rank, '-', len(atoms), atoms.get_global_number_of_atoms())
    atoms.calc = makepot()
    #view(atoms)
    
    dyn = Langevin(atoms, 5*units.fs, temperature_K=0.0, friction=friction)
    logger = MDLogger(dyn, atoms, prefix+'/Melt-phase1.log', stress=True, peratom=True)
    dyn.attach(logger, interval=100)
    set_temp_fric(dyn, atoms)
    unstress = Inhomogeneous_NPTBerendsen(atoms, 50*units.fs, temperature_K=0,
                                          pressure_au=0, taut=500*units.fs,
                                          taup=500*units.fs,
                                          compressibility_au=1/bulkmodulus)
    dyn.attach(unstress.scale_positions_and_cell, interval=10)
    #traj = PickleTrajectory("melting-phase1.traj", "w", atoms)
    fn1 = prefix + "/Melt-phase1.bundle" 
    if master:
        print("Opening file", fn1)
    traj = BundleTrajectory(fn1, "w", atoms, split=True)
    dyn.attach(traj, interval=500)
    
    therm = Thermometer(atoms, prefix+"/TempProfile-phase1.dat")
    dyn.attach(therm.update, interval=500)
    
    for i in range(steps1 // 100):
        dyn.run(100)
        set_temp_fric(dyn, atoms)
    if master:
        print("Closing file", fn1)
    traj.close()
    del traj, atoms
    
    # Reread the bundle
    atoms = BundleTrajectory(fn1).get_atoms(-1, cpulayout2)
    atoms.calc = makepot()
    dyn = VelocityVerlet(atoms, 5*units.fs)
    logger = MDLogger(dyn, atoms, prefix+'/PtMelt-phase2.log', stress=True, peratom=True)
    dyn.attach(logger, interval=1000)  
    unstress = Inhomogeneous_NPTBerendsen(atoms, 50*units.fs,
                                          temperature_K=0,
                                          pressure_au=0,
                                          taup=5000*units.fs, taut=1000*units.fs,
                                          compressibility_au=1/bulkmodulus)
    dyn.attach(unstress.scale_positions_and_cell, interval=10)
    fn2 = prefix + "/Melt-phase2.bundle" 
    if master:
        print("Opening file", fn2)
    traj = BundleTrajectory(fn2, "w", atoms)
    dyn.attach(traj, interval=steps2/100)
    
    therm = Thermometer(atoms, prefix+"/TempProfile-phase2.dat")
    dyn.attach(therm.update, interval=steps2/10)
    
    dyn.run(steps2 - takedata)
    therm2 = AveragingThermometer(atoms)
    dyn.attach(therm2.update, interval=100)
    dyn.run(takedata)  # Run the last part
    T = therm2.get_temp()
    if master:
        print("Melting temperature:", T)
    ReportTest("Melting temperature", T, T_expect, 3)
    if cleanup:
        world.barrier()
        # Cleanup may fail due to the directory not being updated from
        # writes on the other nodes.  Wait for NFS file system.
        time.sleep(60)
        if master:
            shutil.rmtree(prefix)
    world.barrier()
    ReportTest.Summary()    

if __name__ == '__main__':
    main(element = 'Pt',
         T_up = 2000.0,  #K
         T_low = 1000.0, #K
         T_expect = 1357, #K
         bulkmodulus = 230.0 * units.GPa,  # Bulk modulus of Pt
         makepot = EMT
         )



import numpy as np
from asap3 import EMT
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import ReportTest, serial

import pytest

@serial
@pytest.mark.core
def test_largeshear():
    'Test that energy varies smoothly under a large shear.'
    atoms = FaceCenteredCubic(symbol='Cu', size=(4,4,4))
    atoms.calc = EMT()
    uc = atoms.get_cell()
    vals = np.linspace(0, 0.2, 20)

    res = []
    for n,i in enumerate(vals):
        uc[0,2] = i
        atoms.set_cell(uc, scale_atoms=True)
        epot = atoms.get_potential_energy()
        print(n, i, epot)
        res.append(epot)

    poly = np.polyfit(vals, res, 2)
    fits = np.polyval(poly, vals)

    rms = np.sqrt((fits - res) * (fits - res))
    print(rms)
    maxrms = rms.max()
    ReportTest("Worse fit", maxrms, 0.0, 1e-5)



    




#!/usr/bin/env python

"""
Tests that stresses are calculated correctly by Asap

Name: testStress.py

Description: Part of the Asap test suite.  Tests that stresses are
    calculated correctly by calculating various elastic constants from
    strain and stress and comparing them with the same constants
    calculated using strains and energies.  Can also be imported as a
    module, and used to test advanced calculation methods (QC,
    parallel).

Usage: python testStress.py

Expected result: Some elastic constants for Copper and Silver,
followed by the text 'ALL TESTS SUCCEEDED'.

The elastic constants are calculated by straining the crystal in
various modes, and fitting to the energies or the stresses.
Generally, the energies seems to be most sensitive to numerical noise,
and requires a rather large strain interval (1% or above), whereas the
stresses are much less sensitive to this.  On the other hand,
unlinearities influence the stress fits for large strain intervals.  A
strain interval of [-1%, 1%] is a good compromise, where both methods
work.

C11 and C12 calculated 'directly', i.e. in uniaxial strain, is
sensitive to the strain interval when using the energy to fit C11.
Fitting to the stresses work much better.

C11 and C12 can be calculated in an alternative way using a
volume-conserving deformation and fitting to the energies.

All of the above-mentioned calculations are performed.

"""

from asap3 import *
from ase import data, Atoms
from ase.lattice.cubic import Diamond
from ase.lattice.hexagonal import Graphite
from asap3.testtools import ReportTest
import numpy as np

defaultstrains = 0.01 * np.array((-1, -0.75, -0.5, -0.25, 0.0,
                                         0.25, 0.5, 0.75, 1.0))
                                       
book = {}
book['Diamond'] = {"bookbulk": (442.0, 441.4),
                   "bookc11": (1079.0, 1068.9),
                   "bookc12": (124.0, 131.5),
                   "bookc44": (578.0, 735.8),
                   'symbol': 'C'}

# Alexey Bosak and Michael Krisch
# European Synchrotron Radiation Facility, BP 220, F-38043 Grenoble Cedex, France

# Marcel Mohr, Janina Maultzsch, and Christian Thomsen
# Institut für Festkörperphysik, Technische Universität Berlin, Hardenbergstrasse 36, 10623 Berlin, Germany

# Received 22 November 2006; revised 11 January 2007; published 30 April 2007

# The five independent elastic moduli of single-crystalline graphite
# are determined using inelastic x-ray scattering. At room temperature
# the elastic moduli are, in units of GPa, C11=1109, C12=139, C13=0,
# C33=38.7, and C44=4.95. Our experimental results are compared with
# predictions of ab initio calculations and previously reported
# incomplete and contradictory data sets. We obtain an upper limit of
# 1.1 TPa for the on-axis Young’s modulus of homogeneous carbon
# nanotube, thus providing important constraints for further
# theoretical advances and quantitative input to model elasticity in
# graphite nanotubes.

# URL:
# http://link.aps.org.globalproxy.cvt.dk/doi/10.1103/PhysRevB.75.153408
# DOI:
# 10.1103/PhysRevB.75.153408

book['Graphite'] = {"bookbulk": (134.3, 140.77),
                    "bookc11": (1109, 172.75),
                    "bookc12": (139, 124.32),
                    "bookc13": (0, 0),
                    "bookc33": (38.7, 38.7),
                    "bookc44": (75.7, 81.17),
                    'symbol': 'C'}

book['Silicon'] = {"bookbulk": (100.0, 97.8),
                   "bookc11": (124.0, 142.6),
                   "bookc12": (93.7, 75.5),
                   "bookc44": (46.12, 118.9),
                   'symbol': 'Si'}

book['Germanium'] = {"bookbulk": (103.8, 99.27),
                     "bookc11": (124.0, 124.45),
                     "bookc12": (93.7, 86.27),
                     "bookc44": (46.12, 53.44),
                     'symbol': 'Ge'}



defaultstrains = 0.005 * np.array((-1, -0.75, -0.5, -0.25, 0.0,
                                         0.25, 0.5, 0.75, 1.0))

def polynomialLeastSquaresFit(parameters, data, max_iterations=None,
                              stopping_limit = 0.0001):
    """Least-square fit to a polynomial.
    
    Least-squares fit to a polynomial whose order is defined by
    the number of parameter values.

    This is a wrapper function replacing the similar function in
    Scientific.Functions.LeastSquares
    """
    order = len(parameters)
    return (np.polyfit(data[:,0], data[:,1], order)[::-1], None)

def makefits(atoms, strains, indices, shear=0):
    """Do the deformations, and get fits to energies and stresses.

    atoms is a list of atoms.
    strains are the strains as floating point numbers
    indices is a list of indices for the strain components to be used.
        For bulk modulus, it will be ((0,0), (1,1), (2,2)) etc.
    """
    energies = []
    basis = atoms.get_cell()
    vol = np.linalg.det(basis)
    for epsilon in strains:
        if shear:
            adjustment = np.zeros((3,3), float)
            for idx in indices:
                adjustment[idx[0]] += idx[2] * epsilon * basis[idx[1]]
            atoms.set_cell(adjustment + basis, scale_atoms=True)
        else:
            scaling = np.ones((3,3), float)
            for idx in indices:
                scaling[idx[0]] += idx[1]*epsilon
            atoms.set_cell(scaling * basis, scale_atoms=True)
        energy = atoms.get_potential_energy()
        #print ""
        #print epsilon, energy/len(atoms)
        energies.append((epsilon, energy/vol))
    atoms.set_cell(basis, scale_atoms=True)
    energies = np.array(energies)

    #print "Energies:", energies
    energyfit = polynomialLeastSquaresFit((0.0, 0.0, 0.0), energies)
    #print "EnergyFit:", energyfit
    return energyfit

def findlatticeconst(atoms, latticeconstant):
    """Adjust the volume so the atoms have their lowest energy."""
    basis = atoms.get_cell()
    strains = 0.01 * np.array((-0.2, -0.15, -0.1, -0.05, 0.0, 0.05, 0.1,
                               0.15, 0.2))
    for i in range(5):
        energyfit = makefits(atoms, strains,
                             (((0,0), 1), ((1,1), 1), ((2,2),1)))
        pressurefit = -energyfit[0]
        dilation = -pressurefit[1]/(2*pressurefit[2])
        print("Optimizing lattice constant:", latticeconstant, "->", latticeconstant*(1+dilation/3))
        latticeconstant = latticeconstant*(1+dilation/3)
        basis = (1+dilation/3) * basis
        atoms.set_cell(basis, scale_atoms=True)

def elasticconstants(atoms, name, bookbulk, bookc11, bookc12, bookc44,
                     bookc13=None, bookc33=None,
                     fitfact=1.0, fitfact2=1.0):
    """Check the elastic constants."""
    if bookc13 is None:
        bookc13 = bookc12
    if bookc33 is None:
        bookc33 = bookc11
    energyfit = makefits(atoms, defaultstrains,
                         (((0,0), 1), ((1,1), 1), ((2,2),1)))
    bm = 2.0/9.0 * energyfit[0][2] / units.GPa
    print("")
    print("Calculation for", name)
    print("  Bulk modulus from energies:", bm)
    print("  Textbook value:", bookbulk[0])
    print("")
    ReportTest("Bulk modulus (%s, energies)" % name, bm, bookbulk[1],
               0.1*fitfact)
    energyfit = makefits(atoms, defaultstrains, (((0,0),1),))
    c11 = 2.0 * energyfit[0][2] / units.GPa
    energyfit = makefits(atoms, defaultstrains,
                         (((0,0),1), ((1,1), -0.5), ((2,2), -0.5)))
    # C11 - C12
    c11mc12 = 4.0/3.0 * energyfit[0][2] / units.GPa
    c12 = c11 - c11mc12
    print("")
    print("Calculation for", name)
    print("  C_11 from energies:", c11)
    print("  C_12 from energies:", c12)
    print("    Textbook values: C_11 = {:.1f}; C_12 = {:.1f}".format(bookc11[0],
                                                             bookc12[0]))
    
    # B =  (C11 + 2 C12)/3
    altc11 = (3*bm + 2*c11mc12) / 3.0
    altc12 = (3*bm - c11mc12) / 3.0
    print("  C_11 from alternative energies:", altc11)
    print("  C_12 from alternative energies:", altc12)
    print("    Bulk modulus from C_11 and C_12:", (altc11 + 2 * altc12) / 3.0)

    energyfit = makefits(atoms, defaultstrains, (((2,2),1),))
    c33 = 2.0 * energyfit[0][2] / units.GPa
    # C11 - C12
    c11mc12 = 4.0/3.0 * energyfit[0][2] / units.GPa
    c12 = c11 - c11mc12
    print("  C_33 from energies:", c11)
    print(f"    Textbook values: C_33 = {bookc33[0]:.1f}")
    print("")
    
    ReportTest("C11 from energies", c11, bookc11[1], 0.1*fitfact)
    ReportTest("C11 from alt. energies", altc11, c11, 3.0*fitfact)
    ReportTest("C12 from energies", c12, bookc12[1], 0.5*fitfact)
    ReportTest("C12 from alt. energies", altc12, c12, 3.0*fitfact)
    ReportTest("C33 from energies", c33, bookc33[1], 0.1*fitfact)
    print("")

    
    energyfit = makefits(atoms, defaultstrains,
                         (((2,1), (2,2), 1), ((1,2), (1,1), 1)), shear=1)
    c44 = 0.5 * energyfit[0][2] / units.GPa
    print("")
    print("Calculation for", name)
    print("  C_44 from energies:", c44)
    print("  Textbook value:", bookc44[0])
    print("")
    ReportTest("C44 from energies", c44, bookc44[1], 0.1*fitfact)
    print("")
    print("Testing other shear modes:")
    energyfit = makefits(atoms, defaultstrains,
                         (((2,0), (2,2), 1), ((0,2), (0,0), 1)), shear=1)
    c44x = 0.5 * energyfit[0][2] / units.GPa
    ReportTest("C44(alt) from energies", c44x, c44, 0.2*fitfact2)
    energyfit = makefits(atoms, defaultstrains,
                         (((1,0), (1,1), 1), ((0,1), (0,0), 1)), shear=1)
    c44x = 0.5 * energyfit[0][2] / units.GPa
    ReportTest("C44(alt) from energies", c44x, c44, 0.1*fitfact)
    print("")

print_version(1)

for element in ('Diamond', 'Graphite', 'Silicon'):
    symbol = book[element]['symbol']
    print("*** Running test on %s ***" % element)
    z = data.atomic_numbers[symbol]
    if element == 'Graphite':
        latconst = 2.46
        intial = Graphite(directions=[[2,-1,-1,0], [0,1,-1,0], [0,0,0,1]],
                          symbol=symbol, size=(7,7,5),
                          latticeconstant = {'a':2.46,  'c':6.71})
        #view(initial)
        ReportTest("Number of atoms", len(initial), 7*7*5*8, 0)
    else:
        latconst = data.reference_states[z]['a']
        initial = Diamond(directions=[[1,0,0],[0,1,0],[0,0,1]],
                          symbol=symbol, size=(6,6,6))
        ReportTest("Number of atoms", len(initial), 6*6*6*8, 0)
    atoms = Atoms(initial)
    #view(atoms)
    atoms.calc = BrennerPotential()
    atoms.set_momenta(np.zeros((len(atoms),3), float))
    findlatticeconst(atoms, latconst)
    del book[element]['symbol']
    elasticconstants(atoms, element, **book[element])

ReportTest.Summary()



"""Tests if the version numbers are consistent in Python and compiled code."""

from asap3 import _asap, __version__
from asap3 import __file__ as mainfile


versionerror = """


   OOPS  -  BAD ASAP INSTALLATION: INCONSISTENT VERSION NUMBERS

Version number of Python code: %s
Version number of compiled code: %s

Perhaps some modules are loaded from the wrong place.
  Python main module: %s
  Compiled module: %s

Or maybe you just forgot to compile the code after an upgrade.


"""


def check_version(verbose = False):
    "Check if the version numbers are consistent in Python and compiled code."
    try:
        compiled = _asap.get_short_version().strip("'")
    except AttributeError:
        compiled = "unknown (probably 3.0.0)"
    if compiled != __version__:
        try:
            compiledfile = _asap.__file__
        except AttributeError:
            compiledfile = '<built-in>'
        print(versionerror % (__version__, compiled, mainfile, compiledfile))
        raise RuntimeError("Inconsistent Asap version numbers (see message above)")




import sys
import subprocess

def get_proces_output(command):
    qsubproc = subprocess.Popen(command, stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
    (out, err) = qsubproc.communicate()
    errcode = qsubproc.wait()
    if errcode:
        sys.stderr.write(err)
        #sys.stderr.write('Error %i in: %s\n' % (errcode, ' '.join(command)))
        sys.exit()
    return out

def submit_job(command):
    jobid = get_proces_output(command)
    #print 'Job id:', jobid
    if 'asap-qsub' in command:
        jobid = jobid[jobid.find('JOB ID') + 8:-1]
    return jobid

def submit_script(script, filename, command):
    scriptfile = open(filename, 'w')
    scriptfile.write(script)
    scriptfile.close()
    return submit_job(command + [filename])




#!/usr/bin/env python
"""
Tests that EMT works with a twisted, left-hand unit cell when
minimimum-image convention is turned off.
"""

from asap3 import EMT
from asap3.optimize.mdmin import MDMin
from asap3.analysis import CNA, CoordinationNumbers
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest


def checkFCC(atoms, n, name):
    print("Test '%s': %d atoms" % (name, len(atoms)))
    ReportTest((f"Number of atoms ({name})"), len(atoms), n, 0)
    atoms.calc = EMT(minimum_image=False)
    cn = CoordinationNumbers(atoms)
    ReportTest((f"Coordination number is 12 ({name})"),
               np.sum(np.equal(cn, 12)), len(atoms), 0)
    cna = CNA(atoms)
    ReportTest((f"CNA says FCC ({name})"),
               np.sum(np.equal(cna, 0)), len(atoms), 0)
    epot = atoms.get_potential_energy()/len(atoms)
    ReportTest((f"Potential energy ({name})"), epot, 0.0, 1e-3)

@serial
@pytest.mark.core
def test_no_MI_lefthanded():
    directions = [[5,1,3], [3,5,1], [2,2,-7]]
    a = FaceCenteredCubic(directions=directions, size=(2,2,2), symbol="Cu")
    print(a.get_cell())
    checkFCC(a, 5568, "FCC")




from asap3 import EMT, MakeParallelAtoms, BundleTrajectory
from ase import units
import ase.data
from ase.md.verlet import VelocityVerlet
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.lattice.compounds import L1_2
from asap3.mpi import world
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest

cu = ase.data.atomic_numbers['Cu']
au = ase.data.atomic_numbers['Au']

def check_z(a):
    try:
        id = a.get_ids()
    except AttributeError:
        id = np.arange(len(a))
    z = a.get_atomic_numbers()
    z_expect = np.where(id % 4, cu, au)
    assert (z == z_expect).all()

@pytest.mark.slow
def test_bundle_traj_preserves_z(cpulayout, in_tmp_dir):
    cu3au_a = 3.72977
    delete = True

    ismaster = world.rank == 0
    isparallel = world.size != 1

    if ismaster:
        initial = L1_2(size=(15,15,15), symbol=(au,cu),
                    latticeconstant=cu3au_a, pbc=(1,0,0))
    else:
        initial = None
    if isparallel:
        atoms = MakeParallelAtoms(initial, cpulayout)
        print("Min ID", atoms.get_ids().min())
    else:
        atoms = initial.copy()
    
    check_z(atoms)

    print("Simulation: create the bundle")
    # Give a momentum distribution likely to cause migration
    MaxwellBoltzmannDistribution(atoms, temperature_K=5000)
    p = atoms.get_momenta()
    pz = p[10,2]
    p[:,2] += pz

    atoms.calc = EMT()
    if isparallel:
        traj = BundleTrajectory("preservez.bundle", "w", atoms, split=True)
    else:
        traj = BundleTrajectory("preservez.bundle", "w", atoms)
    dyn = VelocityVerlet(atoms, 5*units.fs)
    dyn.attach(traj, interval=50)
    dyn.attach(check_z, interval=25, a=atoms)
    traj.write()
    dyn.run(150)
    traj.close()
    
    print("Reading in serial mode:")
    traj = BundleTrajectory("preservez.bundle")
    for i, atoms in enumerate(traj):
        print(f"Task {world.rank}, step {i}: found {len(atoms)} atoms")
        check_z(atoms)
    traj.close()

    world.barrier()
    print("Reading in parallel mode.")
    traj = BundleTrajectory("preservez.bundle")
    for i in range(len(traj)):
        atoms = traj.get_atoms(i, cpulayout)
        print(f"Task {world.rank}, step {i}: found {len(atoms)} atoms")
        check_z(atoms)
    traj.close()
    del traj

    world.barrier()
    if delete:
        #if ismaster:
        print("Deleting trajectory")
        BundleTrajectory.delete_bundle("preservez.bundle")



# Written by: Rasmus E. Christiansen

# This Class implements an "error" function which gives a measure for how well the seven parameters used in the EMT() calculator are fittet for each element. This is done by comparing a set of quantities to their experimental values and weighting them with regards to the importance of their precision represented by some standard diviation.

# Imported packages
from QuantityCalculatorEMT import Calculator
from asap3.EMT2013 import EMT
import copy
import numpy

# Gobal parameters
parameters = {
#                      E0     s0    V0     eta2    kappa   lambda  n0     
#                      eV     bohr  eV     bohr^-1 bohr^-1 bohr^-1 bohr^-3 
    'H':  numpy.array([-2.21, 0.71, 2.132, 1.652,  2.790,  1.892,  0.00547]),
    'Al': numpy.array([-3.28, 3.00, 1.493, 1.240,  2.000,  1.169,  0.00700]),
    'Cu': numpy.array([-3.51, 2.67, 2.476, 1.652,  2.740,  1.906,  0.00910]),
    'Ag': numpy.array([-2.96, 3.01, 2.132, 1.652,  2.790,  1.892,  0.00547]),
    'Au': numpy.array([-3.80, 3.00, 2.321, 1.674,  2.873,  2.182,  0.00703]),
    'Ni': numpy.array([-4.44, 2.60, 3.673, 1.669,  2.757,  1.948,  0.01030]),
    'Pd': numpy.array([-3.90, 2.87, 2.773, 1.818,  3.107,  2.155,  0.00688]),
    'Pt': numpy.array([-5.85, 2.90, 4.067, 1.812,  3.145,  2.192,  0.00802]),
    'C':  numpy.array([-1.97, 1.18, 0.132, 3.652,  5.790,  2.892,  0.01322]),
    'N':  numpy.array([-4.97, 1.18, 0.132, 2.652,  3.790,  2.892,  0.01222]),
    'O':  numpy.array([-2.97, 1.25, 2.132, 3.652,  5.790,  4.892,  0.00850])}

beta = 1.809        # Calculated from the following equation
                    # beta = ((16 * Pi) / 3)^(1/3) / 2^(1/2)

# The quantities for pure systems that are being used for optimization of the seven parameters are given in the 
# dictionary: quantities.

# The structure of the quantities dictionary is given as follows 
# quantities = {Element: [[BM_exp,BM_sigma,Calc.BulkModulus],
#                    [a0_exp,a0_sigma,Calc.LatticeConstant],
#                    [Evoc_exp,Evoc_sigma,Calc.EvocCalculator],
#                    [EfccEhcp_exp,EfccEhcp_sigma,Calc.EfccEhcpCalculator],
#                    [EfccEbcc_exp,EfccEbcc_sigma,Calc.EfccEbccCalculator],
#                    [C44_exp,C44_sigma,Calc.C44Calculator],
#                    [C11_exp,C11_sigma,Calc.C11_Calculator],
#                    [E_coh_exp,E_coh_sigma,Calc.E_cohCalculator]
#                    [E_surf001_exp,E_surf001_sigma,Calc.Surface001]
#                    [E_surf111_exp,E_surf111_sigma,Calc.Surface001]]
#               ...                    

# The units for the experimental values are given as follows
# [BM_exp]: eV/Å^3 , [a0_exp]: Å , [C44_exp]: eV/Å^3 , [C11_exp]: eV/Å^3 ,
# [Evoc_exp]: eV ,[Ecoh_exp]: eV , [EfccEhcp_exp]: eV , [EfccEbcc_exp]: eV
# [E_surf001_exp]: eV / Å^2, [E_surf111_exp]: eV / Å^2.

# Here XX_exp is the experimental value of the quantity XX.
# XX_sigma is the accepted standard diviation of the quantity XX. 
# BM is the Bulk modulus, a0 the lattice constant, Evoc the vacation energy, EfccEhcp and EfccEbcc the energy 
# difference between a system of atoms of a given element in the FCC and HCP/BCC structures, C44 the Shear modulus,
# C11 another elastic constant, and lastly E_coh the Coherence energy of the system.

# The first column of the dictionary is filled (The experimental values for the parameters)
" FOR NOW THE Evoc_exp or surface energies of Al is not known! This should not be used!! they are all set to Val = 1 in order for the script to run "
# The experimental values are added to the dictionary first.
quantities = {'Al': [[0.474],
                     [4.050],
                     [1], # DONT USE THIS VALUE ITS WRONG!!!!
                     [0.0371],
                     [0.0925],
                     [0.177],
                     [0.666],
                     [3.39],
                     [1], # DONT USE THIS VALUE ITS WRONG!!!!
                     [1]], # DONT USE THIS VALUE ITS WRONG!!!!
              'Cu': [[0.874],
                     [3.615],
                     [1.3],
                     [0.00670],
                     [0.042],
                     [0.472],
                     [1.050],
                     [3.49],
                     [0.1352],
                     [0.1218]],
              'Ag': [[0.624],
                     [4.085],
                     [1.1],
                     [0.00567],
                     [0.0336],
                     [0.288],
                     [0.774],
                     [2.95],
                     [0.07489],
                     [0.07315]],
              'Au': [[1.373],
                     [4.078],
                     [0.9],
                     [0.02], # Uses estimate of EfccEhcp as DFT gives: -0.0201.
                     [0.0174],
                     [0.262],
                     [1.201],
                     [3.81],
                     [0.1015],
                     [0.08008]],
              'Ni': [[1.124],
                     [3.524],
                     [1.6],
                     [0.0325],
                     [0.0678],
                     [0.775],
                     [1.549],
                     [4.44],
                     [0.1514],
                     [0.1255]],
              'Pd': [[1.123],
                     [3.891],
                     [1.4],
                     [0.0363],
                     [0.0410],
                     [0.448],
                     [1.417],
                     [3.89],
                     [0.1452],
                     [0.1198]],
              'Pt': [[1.436],
                     [3.924],
                     [1.5],
                     [0.0825],
                     [0.0951],
                     [0.477],
                     [2.163],
                     [5.84],
                     [0.1706],
                     [0.1435]],
              }

# The procentual values for the standard diviation are converted to be procental values of the experimental values
for Elements,data in quantities.items():
    data[0].append(data[0][0] * 0.035)   # Bulk modulus
    data[1].append(data[1][0] * 0.005)  # Lattice constant
    data[2].append(data[2][0] * 0.05)   # Evoc
    if Elements == 'Au':
        data[3].append(data[3][0]) # STD for Au specifically 
    else:
        data[3].append(data[3][0] * 0.05)   # Hcp-Fcc
    data[4].append(data[4][0] * 0.05)   # Bcc-Fcc
    data[5].append(data[5][0] * 0.035)   # C44
    data[6].append(data[6][0] * 0.035)   # C11
    data[7].append(data[7][0] * 0.05)   # Ecoh
    data[8].append(data[8][0] * 0.1)   # E_surf001
    data[9].append(data[9][0] * 0.1)   # E_surf111

# The calculator objects from the QuantityCalculator package are created and appended with the relavant method to each 
for Elements,data in quantities.items():
    Calc = Calculator(Elements)
    data[0].append(Calc.BulkModulus)         # Bulk modulus
    data[1].append(Calc.LatticeConstant)     # Lattice constant
    data[2].append(Calc.EvocCalculator)      # Evoc
    data[3].append(Calc.EfccEhcpCalculator)  # Hcp-Fcc
    data[4].append(Calc.EfccEbccCalculator)  # Bcc-Fcc
    data[5].append(Calc.C44_Calculator)      # C44
    data[6].append(Calc.C11_Calculator)      # C11
    data[7].append(Calc.E_cohCalculator)     # Ecoh
    data[8].append(Calc.Surface001)          # E_surf001
    data[9].append(Calc.Surface111)          # E_surf111


# # The quantities for alloys used to fit the parameters are presented in quantitiesAlloys 
# quantitiesAlloys = {'AuCu3': [[0.878],       # Bulk modulus
#                               [3.790],       # Lattice constant
#                               [-0.0514]]}    # E_hof

# # The procentual values for the standard diviation are converted to be procental values of the experimental values
# for Elements,data in quantitiesAlloys.items():
#     data[0].append(data[0][0] * 0.035)  # L12_Bulk modulus
#     data[1].append(data[1][0] * 0.010)  # L12_Lattice constant
#     data[2].append(data[2][0] * 0.05)   # L12_HOFE


# # The calculator objects from the QuantityCalculator package are created and appended with the relavant method to each 
# for Elements,data in quantitiesAlloys.items():
#     Calc = Calculator([Elements[0:2],Elements[2:4]])
#     data[0].append(Calc.L12_BulkModulus)     # L12_BulkModulus
#     data[1].append(Calc.L12_LatticeConstant) # L12_LatticeConstant
#     data[2].append(Calc.L12_HOFE)            # L12_HOFE



class ErrorFunction:
    """ This class creates an ErFu object which is used by the MinAlgorithm script in order to optimize a number of  parameters used by the EMT() potential script. """

    def __init__(self,Elements,ElementsAlloys,VariableParameters):

        # The quantities for the elements used in the optimization are pulled from the quantities dictionary.
        self.q = {}
        self.qAlloys = {}
        for i in Elements:
            self.q[i] = quantities[i] 
        for i in ElementsAlloys:
            self.qAlloys[i] = quantitiesAlloys[i]
        # The parameters which are going to be allowed to vary during the optimization process are set.
        self.VP = VariableParameters
        # The initial parametervalues are set and the elements going through optimization are identified
        self.params = copy.deepcopy(parameters)
        self.Elements = Elements
        
    def ErFu(self,ParametersChanged):
        """ This method calculates and returns the value of an "Error" Function based on the current value of the 
            parameters in the input dictionary: ParametersChanged. """
        # The current value of the parameters undergoing optimization in this evaluation 
        # of the Error function are transfered to the used arrays.
        NewParamValues = copy.deepcopy(self.VP)
        ParamNumber = 0
        for i in self.Elements:
            for j in range(7):
                if self.VP[i][j] == 1:
                    NewParamValues[i][j] = ParametersChanged[ParamNumber]
                    ParamNumber += 1

        for i in self.Elements:
            self.params[i] = (self.params[i] - self.params[i] * self.VP[i]) + NewParamValues[i]

        # The calculator is initialized with the current values of the parameters.
        self.EMT = EMT(self.params)
        
        # The ErrorFunction is calculated using the objects given in self.q and self.qAlloys and it value is returned. 
        ErrFunc = 0
        for Elle,List in self.qAlloys.items():
            for i in range(len(List)):
                ErrFunc += (List[i][2](self.EMT,self.params) - List[i][0])**2 / List[i][1]**2
        for Elle,List in self.q.items():
            for i in range(len(List)):
                ErrFunc += (List[i][2](self.EMT,self.params) - List[i][0])**2 / List[i][1]**2
                print("Done something:", ErrFunc)


        # The current parametervalues and Error function value are printed to the user
        for j in self.Elements:
            print(self.params[j])
        print(ErrFunc)

        return ErrFunc













import os
import math
import random
import numpy as np

import pickle
import new

from asap3.nanoparticle_mc.clusteratom import ClusterAtom
from asap3.nanoparticle_mc import data
from ase import Atoms
from ase.data import atomic_numbers, chemical_symbols, reference_states
from ase.lattice.cubic import SimpleCubic, BodyCenteredCubic, FaceCenteredCubic, Diamond
from ase.utils import basestring
from asap3 import FullNeighborList
from asap3 import MonteCarloAtoms

class Cluster(MonteCarloAtoms):
    names = {'numbers':       ('number',       int,   ()  ),
                'positions':     ('position',     float, (3,)),
                'tags':          ('tag',          int,   ()  ),
                'momenta':       ('momentum',     float, (3,)),
                'masses':        ('mass',         float, ()  ),
                'magmoms':       ('magmom',       float, ()  ),
                'charges':       ('charge',       float, ()  ),
                'neighbors':     ('neighbors',    int,   None),
                'coordinations': ('coordination', int,   ()  ),
                'types':         ('type',         int,   ()  ),
               }

    def __init__(self, symbol=None, layers=None, positions=None,
                 latticeconstant=None, symmetry=None, cell=None, 
                 center=None, multiplicity=1, filename=None, debug=0):

        self.debug = debug
        self.multiplicity = multiplicity

        if filename is not None:
            # We skip MonteCarloAtoms.__init__, do it manually.
            self.mc_optim = np.zeros(101, np.intc)
            self.mc_optim[0] = 10000  # MC optim invalid
            self.read(filename)
            return
        
        #Find the atomic number
        if symbol is not None:
            if isinstance(symbol, basestring):
                self.atomic_number = atomic_numbers[symbol]
            else:
                self.atomic_number = symbol
        else:
            raise Warning('You must specify a atomic symbol or number!')

        #Find the crystal structure
        if symmetry is not None:
            if symmetry.lower() in ['bcc', 'fcc', 'hcp']:
                self.symmetry = symmetry.lower()
            else:
                raise Warning('The %s symmetry does not exist!' % symmetry.lower())
        else:
            self.symmetry = reference_states[self.atomic_number]['symmetry'].lower()

        if self.debug:
            print('Crystal structure:', self.symmetry)

        #Find the lattice constant
        if latticeconstant is None:
            if self.symmetry == 'fcc':
                self.lattice_constant = reference_states[self.atomic_number]['a']
            else:
                raise Warning('Cannot find the lattice constant ' +
                               'for a %s structure!' % self.symmetry)
        else:
            self.lattice_constant = latticeconstant

        if self.debug:
            print('Lattice constant(s):', self.lattice_constant)

        #Make the cluster of atoms
        if layers is not None and positions is None:
            layers = list(layers)

            #Make base crystal based on the found symmetry
            if self.symmetry == 'fcc':
                if len(layers) != data.lattice[self.symmetry]['surface_count']:
                    raise Warning('Something is wrong with the defined number of layers!')

                xc = int(np.ceil(layers[1] / 2.0)) + 1
                yc = int(np.ceil(layers[3] / 2.0)) + 1
                zc = int(np.ceil(layers[5] / 2.0)) + 1

                xs = xc + int(np.ceil(layers[0] / 2.0)) + 1
                ys = yc + int(np.ceil(layers[2] / 2.0)) + 1
                zs = zc + int(np.ceil(layers[4] / 2.0)) + 1

                center = np.array((xc, yc, zc)) * self.lattice_constant
                size = (xs, ys, zs)

                if self.debug:
                    print('Base crystal size:', size)
                    print('Center cell position:', center)

                atoms = FaceCenteredCubic(symbol=symbol,
                                          size=size,
                                          latticeconstant=self.lattice_constant,
                                          align=False)

            else:
                raise Warning(('The %s crystal structure is not' +
                               ' supported yet.') % self.symmetry)

            positions = atoms.get_positions()
            numbers = atoms.get_atomic_numbers()
            cell = atoms.get_cell()
        elif positions is not None:
            numbers = [self.atomic_number] * len(positions)
        else:
            numbers = None

        #Load the constructed atoms object into this object
        self.set_center(center)
        MonteCarloAtoms.__init__(self, numbers=numbers, positions=positions,
                                 cell=cell, pbc=False)

        #Construct the particle with the assigned surfasces
        if layers is not None:
            self.set_layers(layers)

    def __repr__(self):
        output = ('Cluster(symbols=%s%s, latticeconstant=%.2f' % 
                  (len(self), chemical_symbols[self.atomic_number], self.lattice_constant))

        for name in self.arrays:
            output += ', %s=...' % name

        output += ', cell=%s' % self._cell.diagonal().tolist()

        if self.get_center() is not None:
            output += ', center=%s' % self.get_center().tolist()

        return output + ')'

    def __getitem__(self, i):
        c = ClusterAtom(atoms=self, index=i)
        return c

    def __delitem__(self, i):
        Atoms.__delitem__(self, i)

        #Subtract one from all neighbor references that is greater than "index"
        if self.has('neighbors'):
            neighbors = self.get_neighbors()
            neighbors = neighbors - (neighbors > i).astype(int)
            self.set_neighbors(neighbors)

    def pop(self, i=-1):
        atom = self[i]
        atom.cut_reference_to_atoms()

        #Subtract one from all neighbor references that is greater than "i"
        if atom.has('neighbors'):
            atom.neighbors = atom.neighbors - (atom.neighbors > i).astype(int)

        del self[i]
        return atom

    def __setitem__(self, i, atom):
        #raise Warning('Use direct assignment like atoms[i].type = x!')

        #If implemented make sure that all values are cleared before copied
        if not isinstance(atom, ClusterAtom):
            raise Warning('The added atom is not a ClusterAtom instance!')

        for name in self.arrays.keys():
            singular, dtype, shape = self.names[name]
            self[i]._set(singular, np.zeros(shape, dtype))

        self[i]._set('number', atom._get('number', True))

        for name in atom._data:
            self[i]._set(name, atom._get(name, True))
        self[i]._mark()

    def append(self, atom):
        if not isinstance(atom, ClusterAtom):
            raise Warning('The added atom is not a ClusterAtom instance!')

        n = len(self)

        for name, a in self.arrays.items():
            b = np.zeros((n + 1,) + a.shape[1:], a.dtype)
            b[:n] = a
            self.arrays[name] = b

        for name in atom._data:
            self[-1]._set(name, atom._get(name, True))

    def extend(self, atoms):
        if not isinstance(atoms, Cluster):
            raise Warning('The added atoms is not a Cluster instance!')

        for atom in atoms:
            self.append(atom)

    def copy(self):
        cluster = Cluster(symbol=self.atomic_number,
                          latticeconstant=self.lattice_constant,
                          symmetry=self.symmetry,
                          cell=self.get_cell(),
                          center=self.get_center())

        for name, a in self.arrays.items():
            cluster.arrays[name] = a.copy()

        return cluster

    #Special Monte Carlo functions to alter the atoms
    def add_atom(self, atom):
        atom.index = len(self)
        self.append(atom)

        if self.has('neighbors'):
            self.update_neighborlist(i=atom.index, new=atom.neighbors)

    def remove_atom(self, i):
        atom = self.pop(i)

        if self.has('neighbors'):
            self.update_neighborlist(old=atom.neighbors)

        return atom

    def move_atom(self, i, other):
        atom = self[i]
        atom.cut_reference_to_atoms()
        self[i] = other

        if self.has('neighbors'):
            self.update_neighborlist(i=i, new=other.neighbors, old=atom.neighbors)

    def update_neighborlist(self, i=None, new=None, old=None):
        """Updates the neighbor list around atoms that is removed and added."""

        neighbor_mapping = data.lattice[self.symmetry]['neighbor_mapping']

        #Add "i" in the new neighbors neighborlists
        if new is not None and i is not None:
            for j, n in enumerate(new):
                if n >= 0 and n != i:
                    self[n].neighbors[neighbor_mapping[j]] = i
                    self[n].coordination += 1
                    self[n].type = self.get_atom_type(self[n].neighbors)
                elif n == i:
                    self[n].neighbors[j] = -1
                    self[n].coordination -= 1
                    self[n].type = self.get_atom_type(self[n].neighbors)

        #Set "-1" in the old neighbors neighborlists
        if old is not None:
            for j, n in enumerate(old):
                if n >= 0:
                    self[n].neighbors[neighbor_mapping[j]] = -1
                    self[n].coordination -= 1
                    self[n].type = self.get_atom_type(self[n].neighbors)

    def make_neighborlist(self):
        """Generate a lists with nearest neighbors, types and coordinations"""
        neighbor_cutoff = data.lattice[self.symmetry]['neighbor_cutoff']
        neighbor_cutoff *= self.lattice_constant
        neighbor_numbers = data.lattice[self.symmetry]['neighbor_numbers']
        neighbor_count = data.lattice[self.symmetry]['neighbor_count']

        get_neighbors = FullNeighborList(neighbor_cutoff, self)

        positions = self.get_positions()
        neighbors = []
        coordinations = []
        types = []

        for i, pos in enumerate(positions):
            nl = get_neighbors[i]
            dl = (positions[nl] - pos) / self.lattice_constant

            neighbors.append([-1] * neighbor_count)
            for n, d in zip(nl, dl):
                name = tuple(d.round(1))
                if name in neighbor_numbers:
                    neighbors[i][neighbor_numbers[name]] = n

            coordinations.append(self.get_atom_coordination(neighbors[i]))
            types.append(self.get_atom_type(neighbors[i]))

        self.set_neighbors(neighbors)
        self.set_coordinations(coordinations)
        self.set_types(types)

    def get_number_of_layers(self, normal):
        """Get the number of layers in the direction given by 'normal'"""
        positions = self.get_positions() - self.get_center()
        
        d = self.lattice_constant * self.get_surface_data('d', normal)
        n = np.array(normal)
        r = np.dot(positions, n) / np.linalg.norm(n)
        layers = np.int(np.round(r.max() / d))

        return layers

    def set_number_of_layers(self, normal, layers):
        """Set the number of layers in the direction given by 'normal'"""
        positions = self.get_positions() - self.get_center()
        
        d = self.lattice_constant * self.get_surface_data('d', normal)
        n = np.array(normal)
        r = np.dot(positions, n) / np.linalg.norm(n)

        actual_layers = self.get_number_of_layers(normal)
        
        if layers > actual_layers:
            if self.debug:
                print ('Expanding the %s plane to %i from %i layers (Not supported yet)'
                       % (normal, layers, actual_layers))
            #raise Warning('Your cluster is not right!')
        
        elif layers < actual_layers:
            if self.debug:
                print ('Cutting the %s plane to %i from %i layers.'
                       % (normal, layers, actual_layers))

            rmax = (layers + 0.5) * d
            mask = np.less(r, rmax)

            for name in self.arrays.keys():
                self.arrays[name] = self.arrays[name][mask]

        else:
            if self.debug:
                print ('Keeping the %s plane at %i layers.'
                       % (normal, layers))

    def get_layers(self):
        surface_names = data.lattice[self.symmetry]['surface_names']
        layers = []

        for n in surface_names:
            layers.append(self.get_number_of_layers(n))

        return np.array(layers, int)

    def set_layers(self, layers):
        surface_names = data.lattice[self.symmetry]['surface_names']
        surface_count = data.lattice[self.symmetry]['surface_count']

        if len(layers) != surface_count:
            raise Warning('The number of surfaces is not right: %i != %i' %
                           (len(layers), surface_count))

        for n, l in zip(surface_names, layers):
            self.set_number_of_layers(normal=n, layers=l)

    def get_diameter(self):
        """Makes an estimate of the cluster diameter based on the average
        distance between opposit layers"""
        surface_mapping = data.lattice[self.symmetry]['surface_mapping']
        surface_names = data.lattice[self.symmetry]['surface_names']
        surface_numbers = data.lattice[self.symmetry]['surface_numbers']
        surface_data = data.lattice[self.symmetry]['surface_data']
        surface_count = data.lattice[self.symmetry]['surface_count']

        d = 0.0
        for s1 in surface_names:
            s2 = surface_names[surface_mapping[surface_numbers[s1]]]
            l1 = self.get_number_of_layers(s1)
            l2 = self.get_number_of_layers(s2)
            dl = surface_data[surface_numbers[s1]]['d'] * self.lattice_constant
            d += (l1 + l2) * dl / surface_count

        return d

    def get_atom_coordination(self, neighbors):
        neighbors = np.array(neighbors)
        neighbors = neighbors[neighbors >= 0]
        return len(neighbors)

    def get_atom_type(self, neighbors):
        neighbors = np.array(neighbors)
        name = tuple((neighbors >= 0).astype(int))

        d = data.lattice[self.symmetry]
        type_numbers = d['type_numbers']
        type_data = d['type_data']

        try:
            return type_data[type_numbers[name]]['type']
        except KeyError:
            return 0

    #Functions to acces the properties
    def get_center(self):
        return self._center.copy()

    def set_center(self, center):
        self._center = np.array(center, float)

    def get_basis(self):
        return self._basis.copy()

    def set_basis(self, basis):
        self._basis = np.array(basis, float)

    def get_positions(self):
        return self.get_array('positions')

    def set_positions(self, positions):
        self.set_array('positions', positions, float)

    def get_neighbors(self):
        return self.get_array('neighbors')

    def get_neighbors_bool(self, index):
        return tuple((self.get_neighbors()[index] >= 0).astype(int))

    def set_neighbors(self, neighbors):
        self.set_array('neighbors', neighbors, int)

    def get_types(self):
        return self.get_array('types')

    def set_types(self, types):
        self.set_array('types', types, int)

    def get_coordinations(self):
        return self.get_array('coordinations')

    def set_coordinations(self, coordinations):
        self.set_array('coordinations', coordinations, int)

    #Functions to store the cluster
    def write(self, filename=None):
        if not isinstance(filename, basestring):
            raise Warning('You must specify a valid filename.')

        if os.path.isfile(filename):
            os.rename(filename, filename + '.bak')

        d = {'symbol': self.atomic_number,
             'latticeconstant': self.lattice_constant,
             'symmetry': self.symmetry,
             'multiplicity': self.multiplicity,
             'center': self.get_center(),
             'cell': self.get_cell(),
             'pbc': self.get_pbc()}

        f = open(filename, 'wb')
        f.write('Cluster')
        pickle.dump(d, f)
        pickle.dump(self.arrays, f)
        f.close()

    def read(self, filename):
        if not os.path.isfile(filename):
            raise Warning('The file specified do not exist.')

        f = open(filename, 'rb')

        try:
            if f.read(len('Cluster')) != 'Cluster':
                raise Warning('This is not a compatible file.')
            d = pickle.load(f)
            self.arrays = pickle.load(f)
        except EOFError:
            raise Warning('Bad file.')

        f.close()

        if 'multiplicity' in d:
            self.multiplicity = d['multiplicity']
        else:
            self.multiplicity = 1

        self.atomic_number = d['symbol']
        self.lattice_constant = d['latticeconstant']
        self.symmetry = d['symmetry']
        self.set_center(d['center'])
        self.set_cell(d['cell'])
        self.set_pbc(d['pbc'])
        self.set_constraint()
        self.adsorbate_info = {}
        self.calc = None

    #Helping fucntions
    def get_surface_data(self, name=None, surface=None):
        if name is None or surface is None:
            return None
        else:
            surface_numbers = data.lattice[self.symmetry]['surface_numbers']
            surface_data = data.lattice[self.symmetry]['surface_data']

            if isinstance(surface, tuple):
                return surface_data[surface_numbers[surface]][name]
            else:
                return surface_data[surface][name]



# Dislocation.py  --  Set up a dislocation with mixed character.

# Copyright (C) 2000-2004 CAMP and Jakob Schiotz <schiotz@fysik.dtu.dk>

"""Module for setting up a dislocation with mixed screw and edge components.

The main function is `Dislocation`, it returns the displacement field
(see `Asap.Setup.DisplacementField`) of a dislocation.  The displacement
field is constructed as the sum of the fields of the edge and screw
components, except if one of the components is zero, in which case it
is omitted.  The displacement field can be applied to a lattice.
"""

from numpy import *
import numbers
from asap3.setup.displacementfield import DisplacementField

__docformat__ = "restructuredtext en"

def magn(v):
    """The magnitude of a vector."""
    return sqrt(dot(v, v))

def Dislocation(origin, line, Burgers, PoissonRatio=None,
                glideplane=None, debug=None):
    """Set up a dislocation with mixed screw and edge character.

    In the following description, a 3-vector may be a tuple, list or
    numeric array with three elements.
    
    This function will set up a straight dislocation through a
    specified point with a given Burgers vector.  The first argument
    ('origin', a 3-vector) specifies a point on the dislocation line,
    and the second argument ('line', a 3-vector) gives the dislocation
    line (only the direction of the vector matters). The third
    parameter ('Burgers', a 3-vector) specifies the Burgers vector of
    the dislocation.  Finally, the fourth argument is an optional
    parameter which specifies Poisson's ratio of the material.  The
    default value is 1/3.
    
    Returns a `setup.displacementfield.DisplacementField` object.
    Displacement fields may be added together (and multiplied by
    scalars).  The dislplacement field can be applied to any 
    Atoms object using field.`apply_to(atoms)`.  The Atoms are 
    modified using the get_positions() and Set_positions() methods.

    Example:

        from asap3 import *
        from ase.lattice.cubic import FaceCenteredCubic
        from asap3.detup.dislocation import Dislocation

        splitting = 5
        size = (30, 25, 7)
        Gold = "Au"

        slab = FaceCenteredCubic(directions=((1,1,-2), (-1,1,0), (1,1,1)),
                                 size=size, symbol=Gold, pbc=False)
        basis = slab.get_cell()
        # Center of system, slight offset to not hit an atom.
        center = 0.5 * array([basis[0,0], basis[1,1], basis[2,2]]) + array([0.1, 0.1, 0.1])
        offset = 0.5 * splitting * slab.miller_to_direction((-1,0,1))

        d1 = Dislocation(center - offset, slab.miller_to_direction((-1,-1,0)),
                         slab.miller_to_direction((-2,-1,1))/6.0)
        d2 = Dislocation(center + offset, slab.miller_to_direction((1,1,0)),
                         slab.miller_to_direction((1,2,1))/6.0)
        atoms = Atoms(slab)
        (d1+d2).apply_to(atoms)
    """

    line = array(line) / magn(line)
    Burgers = array(Burgers)
    screw_b = dot(Burgers, line)
    edge_b = Burgers - screw_b * line
    small = 1e-6
    
    if abs(screw_b) < small:
        result = EdgeDislocation(origin, line, edge_b, PoissonRatio,
                                 debug=debug)
    elif magn(edge_b) < small:
        result = ScrewDislocation(origin, line, screw_b, debug=debug)
    else:
        result = (EdgeDislocation(origin, line, edge_b, PoissonRatio,
                                  debug=debug) +
                  ScrewDislocation(origin, line, screw_b, cut=edge_b,
                                   debug=debug))
    return result


# Helper functions, including specialized functions for defining screw
# and edge dislocation.

def ScrewDislocation(origin, line, b, cut=None, debug=None):
    """Set up a straight screw dislocation.

    Set up a straight screw dislocation through a specified point with
    a given burgers vector.  The first argument ('origin', a 3-tuple)
    specifies a point on the dislocation line, the second ('line', a
    3-tuple) gives the dislocation line (only the direction of the
    vector matters).  The third ('b', a scalar) gives the Burgers
    vector (positive means parallel to the dislocation line, negative
    means antiparallel).  The third optional argument (cut, a 3-tuple)
    gives a direction that lies in the cutting plane when starting in
    the origin.  If not given, (1.0, 0.0, 0.0) is used unless parallel
    to the dislocation line, in which case (0.0, 1.0, 0.0) is used.

    Returns a `setup.displacementfield.DisplacementField` object.
    Displacement fields may be added together (and multiplied by
    scalars).  The dislplacement field can be applied to any 
    Atoms object using field.`apply_to(atoms)`.  The Atoms are 
    modified using the get_positions() and Set_positions() methods.
    """

    # Make sure the burgers vector is a float, not a vector.
    if not isinstance(b, numbers.Real):
        raise TypeError("The 'Burgers vector' argument must be a number (specifying its length).")
    
    # Normalize line and cut, and assure that they are NumPy arrays
    normline = array(line) / magn(line)
    if cut is not None:
        cut = array(cut) / magn(cut)
    else:
        cut = array((1.0, 0.0, 0.0))
        if parallel(normline, cut):
            cut = array((0.0, 1.0, 0.0))
    assert not parallel(normline, cut)
    origin = array(origin)

    # Turn cut to be orthogonal to normline, without changing the plane
    # defined by cut and normline.
    cut = cut - normline * dot(cut, normline)
    cut = cut / magn(cut)
    return DisplacementField(lambda r, f=_screw, origin=origin, line=normline,
                             b=b, cut=cut, debug=debug:
                             f(r,origin,line,b,cut,debug))

def _screw(r, origin, line, b, cut, debug):
    """The screw dislocation displacement field."""
    if debug:
        print("r: ", r)
        print("origin: ", origin)
    r = r - origin
    if debug:
        print("r: ", r)
    x = vdot(r, cut)
    y = vdot(r, cross(line, cut))
    z = vdot(r, line)         # Distance along line
    if debug:
        print("x: ", x)
        print("y: ", y)
        print("z: ", z)
    x = where(equal(x, 0) * equal(y, 0), ones(x.shape, float), x)
    theta = arctan2(y, x)
    if debug:
        print("theta: ", theta)
    return line * (b * theta / (2 * pi))[:,newaxis]

def EdgeDislocation(origin, line, Burgers, PoissonRatio=None, debug=None):
    """Sets up a straight edge dislocation.

    Sets up a straight edge dislocation through a specified point with
    a given Burgers vector.  The first argument ('origin', a 3-tuple)
    specifies a point on the dislocation line, and the second argument
    ('line', a 3-tuple) gives the dislocation line (only the direction of the
    vector matters). The third parameter
    ('Burgers', a 3-tuple) specifies the Burgers vector of the dislocation.
    The Burgers vector must be perpendicular to the line vector. Finally, the
    fourth argument is an optional parameter which specifies Poisson's ratio
    (unlike in the case of a screw dislocation, this number is needed in order
    to calculate the displacement field). The default value of Poisson's ratio
    is 1/3. 

    Returns a `setup.displacementfield.DisplacementField` object.
    Displacement fields may be added together (and multiplied by
    scalars).  The dislplacement field can be applied to any 
    Atoms object using field.`apply_to(atoms)`.  The Atoms are 
    modified using the get_positions() and set_positions() methods.
    
    """

    # Normalize line and Burgers, and check their orthogonality
    b = magn(Burgers)
    normBurgers = array(Burgers) / b
    normline = array(line) / magn(line)
    assert perp(normline, normBurgers) 
    origin = array(origin)
    if PoissonRatio is None:
        PoissonRatio=1.0/3.0
    return DisplacementField(lambda r, f=_edge, origin=origin, line=normline,
                             Burgers=normBurgers, b=b,
                             PoissonRatio=PoissonRatio, debug=debug:
                             f(r,origin,line,Burgers,b,PoissonRatio,debug))

def _edge(r, origin, line, Burgers, b, PoissonRatio, debug):
    """The edge dislocation displacement field."""
    if debug:
        print("r: ", r)
        print("origin: ", origin)
    r = r - origin 
    x = vdot(r, Burgers) 
    y = vdot(r, cross(line, Burgers)) 
    z = vdot(r, line) 
    if debug:
        print("x: ", x) 
        print("y: ", y)
    theta = arctan2(y, x)
    l = x*x+y*y
    epsilon = 0.1 * b**2
    ux = b / (2 * pi) * (theta + cos(theta) * sin(theta) / (2*(1 - PoissonRatio)))
    uy = b / (2 * pi * 4 * (1 - PoissonRatio)) * ((1-2*PoissonRatio)*log(l+epsilon) + cos(2*theta))
    if debug:
        print("theta: ", theta)
        print("ux: ", ux)
        print("ux: ", uy)
    return  Burgers * ux[:,newaxis]- cross(line, Burgers) * uy[:,newaxis]
    


def vdot(a, b):
    """Dot product of an array of vectors with a vector."""
    # a has shape (N, 3), b has shape (3,) or (N, 3)
    assert a.shape[-1] == 3 and b.shape[-1] == 3
    at = transpose(a)
    bt = transpose(b)
    return at[0] * bt[0] + at[1] * bt[1] + at[2] * bt[2]
    
def magn(v):
    """The magnitude of a vector."""
    return sqrt(dot(v, v))

def cross(a, b):
    """The cross product of two vectors."""
    return array((a[1]*b[2] - b[1]*a[2],
                  a[2]*b[0] - b[2]*a[0],
                  a[0]*b[1] - b[0]*a[1]))

def perp(v1, v2, small=1e-6):
    """Tests if two vectors are perpendicular."""
    if dot(v1,v2) < small:
        return 1
    else:
        return 0


def parallel(v1, v2, small=1e-6):
    """Tests if two vectors are parallel."""
    c = cross(v1, v2)
    if dot(c, c) < small:
        return 1
    else:
        return 0
    



"""Holds functions to make plotting of coordinations easier."""
import pickle
import os
import sys
import numpy as np
import warnings
class MC_Result:

	def __init__(self,av_coordi=None,av_energy=None,tgas=None,pgas=None,tamc=None,
	natoms=None,diameter=None,filename=None):
		self.version = 1.0
		self.av_coordi = av_coordi
		self.av_energy = av_energy
		self.natoms = natoms
		self.tgas = tgas
		self.pgas = pgas
		self.tamc = tamc
		self.diameter = diameter
		self.filename = filename
		
		
	def PickleIt(self):
		"""Saves object to pickle, parameters should have been set before"""
		
		if self.filename ==None:
			op = "./result0_N"+str(self.natoms)+".mcresult"
			i=1
			while os.path.exists(op):
				op = "./result"+str(i)+"_N"+str(self.natoms)+".mcresult"
			
			self.filename = op
			
		f = open(self.filename, "wb")
		pickle.dump(self,f)
		
		f.close()
		
		
	def LoadPickle(self,filename):
		
		f = open(filename, "rb")
		p = pickle.load(f)
		f.close()
		if p.version != self.version:
			warnings.warn("Version used does not match pickled version")
		self.av_coordi = p.av_coordi
		self.av_energy = p.av_energy
		self.natoms = p.natoms
		self.tgas = p.tgas
		self.pgas = p.pgas
		self.diameter = p.diameter
		self.tamc = p.tamc
		self.filename = p.filename
	

		
		
		



import numpy as np
from asap3.test.nvt_module import main
from asap3.md.npt import NPT
from asap3.md.nvtberendsen import NVTBerendsen

from asap3.test.pytest_markers import serial, parallel
from ase.parallel import world

import pytest

@pytest.mark.slow
@serial
def test_nvt_oldnpt():
    def nptdynmaker(atoms, T0, dt, tau, rng, logint):
        return NPT(atoms, timestep=dt, temperature_K=T0, ttime=tau, logfile='-', loginterval=logint)

    def berdynmaker(atoms, T0, dt, tau, rng, logint):
        return NVTBerendsen(atoms, timestep=dt, temperature_K=T0, taut=tau, logfile='-', loginterval=logint)

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()
    main(nptdynmaker, rng, initdyn=berdynmaker, failfluct=True)



from asap3 import EMT, print_version
from asap3.analysis import CoordinationNumbers
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
import ase.data
from asap3.test.pytest_markers import ReportTest, serial

import pytest

element = "Cu"
latconst = ase.data.reference_states[ase.data.atomic_numbers[element]]['a']
nndist = latconst / np.sqrt(2)

sizes = [(3,3,3), (2, 2, 2), (1, 1, 1)]
shells = (1, 2, 3, 4)
expected = (12, 18, 42, 54)
withpotential = (False, True)

@serial
@pytest.mark.core
@pytest.mark.parametrize('size', sizes)
@pytest.mark.parametrize('usepot', withpotential)
def test_cn(size, usepot):
    print(f'Size: {size}  -  use potential: {usepot}')
    atoms = FaceCenteredCubic(
        directions=[[1,0,0],[0,1,1],[0,0,1]], 
        size=size,
        symbol=element
    )
    
    if usepot:
        atoms.calc = EMT()
        epot = atoms.get_potential_energy()

    for shell, expect in zip(shells, expected):
        cutoff = nndist * (np.sqrt(shell) + np.sqrt(shell + 1)) / 2
        cn = CoordinationNumbers(atoms, cutoff)
        print(cn)
        assert np.equal(cn, expect).all()




"""Usage: python test_reproducible.py [init]

Used to test if the EMT potential still produces the same energies, forces
and stresses as it used to.

Call it the first time with the "init" argument to generate a data file.

Subsequent calls calculate the same quantities, but compare them with the
database.

This is the very first test introduced in Asap, going all the way back to 
Asap version 1.  The script has been updated to handle changes in syntax,
and the input file has been converted as we moved from Numeric to
numarray to numpy.
"""

import pytest

from numpy import sin, arange
from asap3 import EMT, Atoms
import pickle, os, sys, time

from asap3.test.pytest_markers import serial

@pytest.mark.core
@serial
def test_reproducible(datadir):
    datafile = datadir / "test_reproducible_results.pickle"
    infile = datadir / "test_reproducible_in.pickle"

    with open(infile, 'rb') as inpickle:
       indata = pickle.load(inpickle, encoding='latin1')
    lattice = Atoms(positions=indata['positions'], cell=indata['cell'], symbols=["Cu"]*len(indata['positions']), pbc=True)
    
    print("Number of atoms", len(lattice))
    print(lattice.get_cell())
    # Perturb the positions
    r = lattice.get_positions()
    dr =  0.05 * sin(arange(3*len(lattice)))
    dr.shape = (-1,3)
    r += dr
    lattice.set_positions(r)

    atoms1 = Atoms(lattice)
    atoms1.calc = EMT()

    print("Total energy:", atoms1.get_potential_energy())
    print("Stress:", atoms1.get_stress(include_ideal_gas=True))

    z = lattice.get_atomic_numbers()
    z[0] = z[1] = 47
    lattice.set_atomic_numbers(z)
    atoms2 = Atoms(lattice)
    atoms2.calc = EMT()
    
    with open(datafile, "rb") as datapickle:
        data = pickle.load(datapickle, encoding='latin1')
    Cu = data["Cu"]
    AgCu = data["AgCu"]
    print("*** Checking pure copper ***")
    e = 0
    e = e + evaluate("energies", Cu["energies"], atoms1.get_potential_energies())
    e = e + evaluate("forces", Cu["forces"], atoms1.get_forces())
    e = e + evaluate("stresses", Cu["stresses"], atoms1.get_stresses())
    print("*** Checking silver in copper ***")
    e = e + evaluate("energies", AgCu["energies"], atoms2.get_potential_energies())
    e = e + evaluate("forces", AgCu["forces"], atoms2.get_forces())
    e = e + evaluate("stresses", AgCu["stresses"], atoms2.get_stresses())
    if e == 6:
        print("*** All tests passed ***")
    else:
        raise RuntimeError("*** THERE WERE ERRORS IN SOME TESTS! ***")

def evaluate(text, expected, actual):
    diff = max(abs(expected.flat[:] - actual.flat[:]))
    passed = diff < 1e-10
    assert passed, f"Checking {text}: max error = {diff:g}"
    return passed

    




from asap3.analysis.localstructure import CoordinationNumbers, CNA, \
     RestrictedCNA, FullCNA
from asap3.analysis.ptm import PTM, PTMobserver, PTMdislocations
from asap3.analysis.rdf import RadialDistributionFunction
from asap3.analysis.clusterposition import ClusterCenter



from ase.filters import StrainFilter
from ase.optimize import MDMin
from asap3 import EMT, Langevin, Atoms, units
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest

if getattr(Atoms, '_ase_handles_dynamic_stress', False):
    stresshack = {'include_ideal_gas': True}
else:
    stresshack = {}

@serial
def test_strainfilter():
    size = 5
    atoms = FaceCenteredCubic(size=(size,size,size), symbol="Cu", pbc=True)
    defsize = atoms.get_volume()
    atoms.set_cell(atoms.get_cell() * 1.1, scale_atoms=True)
    atoms.calc = EMT()

    def printvol(a):
        print("Volume:", a.get_volume(), " energy:",\
            atoms.get_potential_energy() + atoms.get_kinetic_energy(),\
            " stress:", atoms.get_stress(include_ideal_gas=True)[:3])

    f = StrainFilter(atoms, [1, 1, 1, 0, 0, 0], include_ideal_gas=True)
    opt = MDMin(f, logfile="/dev/null", dt=0.01/(atoms.get_cell()[0,0]))
    printvol(atoms)
    opt.attach(printvol, 10, atoms)
    opt.run(0.01)
    printvol(atoms)
    print("Original vol:", defsize)
    print()
    stress = atoms.get_stress(include_ideal_gas=True)
    for i in range(6):
        ReportTest("Stress component %d (T=0)" % (i,), stress[0], 0.0, 1e-5)

    atoms = FaceCenteredCubic(size=(size,size,size), symbol="Cu", pbc=True)
    atoms.calc = EMT()
    dyn = Langevin(atoms, 10*units.fs, temperature_K=500, friction=0.02)
    dyn.attach(printvol, 100, atoms)
    dyn.run(500)

    print()

    f = StrainFilter(atoms, [1, 1, 1, 0, 0, 0], include_ideal_gas=True)
    opt = MDMin(f, logfile="/dev/null", dt=0.01/(atoms.get_cell()[0,0]))
    printvol(atoms)
    opt.attach(printvol, 10, atoms)
    opt.run(0.01)
    printvol(atoms)
    print("Original vol:", defsize)
    print()

    stress = atoms.get_stress(include_ideal_gas=True)
    for i in range(6):
        ReportTest("Stress component %d (T=500K)" % (i,), stress[0], 0.0, 1e-5)



from asap3 import EMT, LennardJones, EMT2013
from ase.lattice.cubic import FaceCenteredCubic
from asap3.test.pytest_markers import ReportTest
from asap3.EMT2013Parameters import sihb_PtY_parameters
from asap3.test.pytest_markers import serial

import pytest

def MakeLJ():
    "Lennard-Jones potential."
    p = LennardJones([78], [0.15], [2.35], -1.0, False)
    return p

def MakeEMT2013():
    "EMT2013 potential."
    return EMT2013(sihb_PtY_parameters)

@serial
@pytest.mark.core
@pytest.mark.parametrize('pot', (EMT, MakeLJ, MakeEMT2013), ids=('EMT', 'LJ', 'EMT2013'))
def test_changeatoms(pot):
    print("Testing:", pot.__doc__.split("\n")[0])
    
    atoms1 = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              size=(6,6,6), symbol="Pt")
    emt = pot()
    atoms1.calc = emt
    f1 = atoms1.get_forces()


    atoms2 = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              size=(5,5,5), symbol="Pt")
    atoms2.calc = emt
    f3 = atoms1.get_forces()
    maxdev = abs(f3 - f1).max()
    print(maxdev)
    ReportTest("Max error 1:", maxdev, 0.0, 1e-6)

    atoms1 = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              size=(6,6,6), symbol="Pt")
    emt = pot()
    atoms1.calc = emt
    f1 = atoms1.get_forces()


    atoms2 = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              size=(5,5,5), symbol="Pt")
    atoms2.calc = emt
    atoms3 = atoms2.copy()
    atoms3.calc = pot()
    f0 = atoms3.get_forces()
    f2 = atoms2.get_forces()
    f3 = atoms1.get_forces()
    maxdev = abs(f2 - f0).max()
    print(maxdev)
    ReportTest("Max error 2:", maxdev, 0.0, 1e-6)
    maxdev = abs(f3 - f1).max()
    print(maxdev)
    ReportTest("Max error 3:", maxdev, 0.0, 1e-6)



# Class which creates "Calculator" objects that uses the EMT() model to calculate the quantities used in the ErrorFunction script in order to optimize the value of the seven parameters used in the EMT() model. 

# Written by: Rasmus E. Christiansen

# Imported Packages
import numpy
from asap3.EMT2013 import EMT
from ase.lattice.cubic import FaceCenteredCubic, BodyCenteredCubic
from ase.lattice.hexagonal import HexagonalClosedPacked
from ase.lattice.compounds import L1_0,L1_2
from ase.data import atomic_numbers, chemical_symbols 
from ase.units import Bohr
import numpy.lib.polynomial as numpyfit
from ase.utils.eos import EquationOfState
from ase.all import view
from ase.optimize import BFGS
from ase.io import read

# Global parameters
parameters = {
#          E0     s0    V0     eta2    kappa   lambda  n0      Lattice type
#          eV     bohr  eV     bohr^-1 bohr^-1 bohr^-1 bohr^-3 
    'H':  (-2.21, 0.71, 2.132, 1.652,  2.790,  1.892,  0.00547, 'dimer'),
    'Al': (-3.28, 3.00, 1.493, 1.240,  2.000,  1.169,  0.00700, 'fcc'),
    'Cu': (-3.51, 2.67, 2.476, 1.652,  2.740,  1.906,  0.00910, 'fcc'),
    'Ag': (-2.96, 3.01, 2.132, 1.652,  2.790,  1.892,  0.00547, 'fcc'),
    'Au': (-3.80, 3.00, 2.321, 1.674,  2.873,  2.182,  0.00703, 'fcc'),
    'Ni': (-4.44, 2.60, 3.673, 1.669,  2.757,  1.948,  0.01030, 'fcc'),
    'Pd': (-3.90, 2.87, 2.773, 1.818,  3.107,  2.155,  0.00688, 'fcc'),
    'Pt': (-5.85, 2.90, 4.067, 1.812,  3.145,  2.192,  0.00802, 'fcc'),
    'C':  (-1.97, 1.18, 0.132, 3.652,  5.790,  2.892,  0.01322, 'dimer'),
    'N':  (-4.97, 1.18, 0.132, 2.652,  3.790,  2.892,  0.01222, 'dimer'),
    'O':  (-2.97, 1.25, 2.132, 3.652,  5.790,  4.892,  0.00850, 'dimer')}

beta = 1.809    # Calculated from the following equation
                # beta = ((16 * Pi) / 3)^(1/3) / 2^(1/2)

class Calculator:
    """ The class creates calculator objects which can be used to calculate the following 16 different quantities for 
        a crystal in the FCC structure using the new version of the EMT calculator. For pure crystals: Bulk modulus, 
        Lattice Constant, Vacancy formation energy, Energy difference between the element in question in an FCC 
        configuration and an HCP or BCC configuration, The elastic constants C44 and C11 along with the cohesive energy 
        and the surface energy for a 001 and 111 surface. 
        For the L10 and L12 alloys of two elements: The heat of formation energy, the bulk modulus and the lattice 
        constant for the L10 and L11 alloys.  """
    def __init__(self,Element,Size=3,Tol=0.002):
        # The element chosen for the calculator is set
        self.Element = Element
        self.Size = Size
        self.Tol = Tol

    def BM_LC_Calculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ The method BM_LC_Calculator uses the EMT calculator to find the lattice constant which gives 
            the lowest possible energy for a system of atoms and from a polynomial fit of the energy as a function of 
            the lattice constant it calculates and returns the Bulk modulus, the volume of the system at the lowest 
            possible energy and this energy """
        # Three values for the lattice constant, a0, are chosen. The highest and lowest value is chosen so that it     
        # is certain that these are to high/low respectively compared to the "correct" value. This is done by using the 
        # experimental value of the lattice constant, a0_exp, as the middle value and to high/low values are then: 
        # a0_exp +- a0_mod, a0_mod = a0_exp/10
        a0_exp = numpy.sqrt(2) * beta * PARAMETERS[self.Element][1] * Bohr
        a0_mod = a0_exp * 0.10 
        a0_guesses = numpy.array([a0_exp - a0_mod,a0_exp,a0_exp + a0_mod])
    
        # An atoms object consisting of atoms of the chosen element is initialized 
        atoms = FaceCenteredCubic(size=(self.Size,self.Size,self.Size), symbol=self.Element)
        atoms.set_calculator(EMT)

        # An identity matrix for the system is saved to a variable
        IdentityMatrix = numpy.array([[1,0,0],[0,1,0],[0,0,1]])

        # An array for the energy for the chosen guesses for a0 is initialized:
        E_guesses = numpy.zeros(5)

        # The energies are calculated
        for i in range(3):
        # Changes the lattice constant for the atoms object
            atoms.set_cell(a0_guesses[i] * self.Size * IdentityMatrix,scale_atoms=True)
            # Calculates the energy of the system for the new lattice constant
            E_guesses[i] = atoms.get_potential_energy()
           
        # Bisection is used in order to find a small interval of the lattice constant for the minimum of the energy (an 
        # abitrary interval length is chosen), This is possible because we are certian from theory that there is only 
        # one minimum of the energy function in the interval of interest and thus we wont "fall" into a local minimum 
        # and stay there by accident.
        while (a0_guesses[2]-a0_guesses[0]) >= self.Tol:
            if min([E_guesses[0],E_guesses[2]]) == E_guesses[0]:
                # A new guess for the lattice constant is introduced
                a0_new_guess = 0.67 * a0_guesses[1] + 0.33 * a0_guesses[2]
                # The energy for this new guess is calculated
                atoms.set_cell(a0_new_guess * self.Size * IdentityMatrix,scale_atoms=True)
                E_new_guess = atoms.get_potential_energy()
                # A check for changes in the energy minimum is made and the guesses for a0 and their corrosponding
                # energies are ajusted.
                if min(E_new_guess,min(E_guesses[0:3])) != E_new_guess:
                    a0_guesses[2] = a0_new_guess
                    E_guesses[2] = E_new_guess 
                else:
                    a0_guesses[0] = a0_guesses[1]
                    a0_guesses[1] = a0_new_guess
                    E_guesses[0] = E_guesses[1]
                    E_guesses[1] = E_new_guess
         

            elif min([E_guesses[0],E_guesses[2]]) == E_guesses[2]: 
                # A new guess for the lattice constant is introduced
                a0_new_guess = 0.33 * a0_guesses[0] + 0.67 * a0_guesses[1]
                # The energy for this new guess is calculated
                atoms.set_cell(a0_new_guess * self.Size * IdentityMatrix,scale_atoms=True)
                E_new_guess = atoms.get_potential_energy()
                # A check for changes in the energy minimum is made and the guesses for a0 and their corrosponding
                # energies are ajusted.
                if min(E_new_guess,min(E_guesses[0:3])) != E_new_guess:
                    a0_guesses[0] = a0_new_guess
                    E_guesses[0] = E_new_guess 
                else:
                    a0_guesses[2] = a0_guesses[1]
                    a0_guesses[1] = a0_new_guess
                    E_guesses[2] = E_guesses[1]
                    E_guesses[1] = E_new_guess

        
        # An estimate of the minimum energy can now be found from a second degree polynomial fit through the three 
        # current guesses for a0 and the corresponding values of the energy.
        Poly = numpyfit.polyfit(a0_guesses,E_guesses[0:3],2)

        # The lattice constant corresponding to the lowest energy from the Polynomiel fit is found
        a0 = - Poly[1] / (2 * Poly[0])

        # Now five guesses for a0 and the corresponding energy are evaluated from and around the current a0.
        a0_guesses = a0 * numpy.array([1 - 2 * self.Tol / 5,1 - self.Tol / 5,1,1 + self.Tol / 5,1 + 2 * self.Tol / 5])


        for i in range(5):
            # Changes the lattice constant for the atoms object
            atoms.set_cell(a0_guesses[i] * self.Size * IdentityMatrix,scale_atoms=True)
            # Calculates the energy of the system for the new lattice constant
            E_guesses[i] = atoms.get_potential_energy()

        # The method EquationOfState is now used to find the Bulk modulus and the minimum energy for the system.

        # The volume of the sample for the given a0_guesses
        Vol = (self.Size * a0_guesses )**3
        
        # The equilibrium volume, energy and bulk modulus are calculated
        (Vol0 , E0 , B ) = EquationOfState(Vol.tolist(),E_guesses.tolist()).fit()

        return (Vol0,E0,B,Vol0**(1./3) / self.Size)

    def BulkModulus(self,EMT,PARAMETERS):
        return self.BM_LC_Calculator(EMT,PARAMETERS)[2]

    def LatticeConstant(self,EMT,PARAMETERS):
        return self.BM_LC_Calculator(EMT,PARAMETERS)[3]

    def EvocCalculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ This method calculates the vacancy formation energy for the system of atoms. That is, this method calculates 
            the potential energy for a complete system of atoms, then removes an atom and calculates the potential energy
            again. Then the energy for the full system (scaled with the number of atoms in the reduced system) are  
            subtracted from that of the reduced system and the vacancy formation energy, Evoc, is returned. """
        # The atoms object is initialized for the chosen size and type of system
        atoms = FaceCenteredCubic(size=(self.Size,self.Size,self.Size), symbol=self.Element)
        # The EMT calculator given is attached to the atoms object
        atoms.set_calculator(EMT)
        
        # The energy of the full system is calculated
        E_FullSystem = atoms.get_potential_energy()
        
        # an atom is removed from the system
        atoms.pop()
        
        # The energy of the reduced system is calculated
        E_ReducedSystem = atoms.get_potential_energy()

        # The energy of a full system compared to the energy pr atom of the reduced system is calculated and returned
        return E_ReducedSystem - E_FullSystem * len(atoms) / (len(atoms) + 1)

    def EfccEhcpCalculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ This method uses the EMT calculator to calculate and return the difference in energy between a system of 
            atoms placed in the HCP and FCC structure. """            
        # The atoms objects are created using the input size and element and the energy calculator 
        # is set to the EMT calculator
        # The Lattice Constants, a,c, for the HCP lattice is here given by the nearest neighbor distance of 
        # the system in the FCC crystal structure, a = dnn, and the ideal relation between a and 
        # c: a/c = sqrt(8/3) => c = dnn / sqrt(8/3)
        a = beta * PARAMETERS[self.Element][1] * Bohr 
        c = a * numpy.sqrt(8./3.)
        
        # The HCP crystal is created, the size of the crystal is defined as 5,5,5, any smaller crystal will result in 
        # Neighborlist errors.
        atoms1 = HexagonalClosedPacked(size=(5,5,5),directions=[[2,-1,-1,0],[0,1,-1,0],[0,0,0,1]], 
                 symbol=self.Element,latticeconstant={'a':a,'c':c})
        atoms1.set_calculator(EMT)
        
        # The FCC crystal is created
        atoms2 = FaceCenteredCubic(size=(self.Size,self.Size,self.Size),symbol=self.Element)
        atoms2.set_calculator(EMT)

        # The energy difference pr atom is calculated and returned
        return atoms1.get_potential_energy() / len(atoms1) - atoms2.get_potential_energy() / len(atoms2)

    def EfccEbccCalculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ This method uses the given EMT calculator to calculate and return the difference in energy between a system 
            of atoms placed in the BCC and FCC structure. """ 
        # The atoms objects are created using the input size and element and the energy calculator 
        # is set to the EMT calculator
        # The Lattice Constant for the BCC lattice is here given so that the volume pr atom of the system is 
        # held fixed for the FCC and BCC lattice, that is Vol_pr_atom_BCC = Vol_pr_atom_FCC.
        atoms1 = FaceCenteredCubic(size=(self.Size,self.Size,self.Size),symbol=self.Element)
        atoms1.set_calculator(EMT)
        # The Lattice constant which produces the same volume pr atom for an BCC crystal is calculated
        LCBcc = 1./(2.**(1./3.)) * beta * PARAMETERS[self.Element][1] * Bohr * numpy.sqrt(2)
        atoms2 = BodyCenteredCubic(size=(self.Size,self.Size,self.Size),symbol=self.Element,latticeconstant=LCBcc)
        atoms2.set_calculator(EMT)
        
        # The energy difference pr atom is calculated and returned
        return atoms2.get_potential_energy() / len(atoms2) - atoms1.get_potential_energy() / len(atoms1)

    def C44_Calculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ This method uses the given EMT calculator to calculate and return the value of the matrix element C44 for 
            a system of atoms of a given element type. The calculation is done by using that:
            C44 = 1 / Volume * d^2/depsilon^2 (E_system) where epsilon is the displacement in one direction of the 
            system along one axis diveded by the highth of the system. """

        # An atom object is created and the calculator attached
        atoms = FaceCenteredCubic(size=(self.Size,self.Size,self.Size),symbol=self.Element)
        atoms.set_calculator(EMT)

        # The volume of the sample is calculated
        Vol = atoms.get_volume()    

        # The value of the relative displacement, epsilon, is set 
        epsilon = 1. / 1000

        # The matrix used to change the unitcell by n*epsilon is initialized
        LMM = numpy.array([[1,0,-10*epsilon],[0,1,0],[0,0,1]])
        
        # The original unit cell is conserved
        OCell = atoms.get_cell()

        # The array for storing the energies is initialized
        E_calc = numpy.zeros(20)

        # The numerical value of C44 is calculated
        for i in range(20):
            # The new system cell based on the pertubation epsilon is set
            atoms.set_cell(numpy.dot(OCell, LMM), scale_atoms=True)
            # The energy of the system is calculated
            E_calc[i] = atoms.get_potential_energy()
            # The value of LMM is updated
            LMM[0,2] += epsilon
            
        # A polynomial fit is made for the energy as a function of epsilon 

        # The displaced axis is defined
        da = numpy.arange(-10, 10) * epsilon
        
        # The fit is made
        Poly = numpyfit.polyfit(da,E_calc,2)

        # Now C44 can be estimated from this fit by the second derivative of Poly = a * x^2 + b * x + c , multiplied
        # with 1 / (2 * Volume) of system
        C44 = 2. / Vol * Poly[0]
        
        return C44

    def C11_Calculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ This method uses the given EMT calculator to calculate and return the value of the matrix element C11 for 
            a system of atoms of a given element type. The calculation is done by using that:
            C11 = 1 / Volume * d^2/depsilon^2 (E_system). """
        # An atom object is created and the calculator attached
        atoms = FaceCenteredCubic(size=(self.Size,self.Size,self.Size),symbol=self.Element)
        atoms.set_calculator(EMT)

        # The volume of the sample is calculated
        Vol = atoms.get_volume()    

        # The value of the relative displacement, epsilon, is set 
        epsilon = 1. / 1000

        # The matrix used to change the unitcell by n*epsilon is initialized
        LMM = numpy.array([[1-10*epsilon,0,0],[0,1,0],[0,0,1]])
        
        # The original unit cell is conserved
        OCell = atoms.get_cell()

        # The array for storing the energies is initialized
        E_calc = numpy.zeros(20)

        # The numerical value of C11 is calculated
        for i in range(20):
            # The new system cell based on the pertubation epsilon is set
            atoms.set_cell(numpy.dot(OCell, LMM), scale_atoms=True)
            # The energy of the system is calculated
            E_calc[i] = atoms.get_potential_energy()
            # The value of LMM is updated
            LMM[0,0] += epsilon
        

        # A second degree polynomial fit is made for the energy as a function of epsilon 

        # The displaced axis is defined
        da = numpy.arange(-10, 10) * epsilon
        
        # The fit is made
        Poly = numpyfit.polyfit(da,E_calc,2)
        
        # Now C11 can be estimated from this fit by the second derivative of Poly = a * x^2 + b * x + c, multiplied
        # with 1 / (2 * Volume) of system
        C11 = 2. / Vol * Poly[0]
        
        return C11

    def E_cohCalculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ Calculates the Cohesive energy of a system of atoms using the EMT calculator specified. """

        # As the EMT calculator calculates the energy of the system such that the energy of the individual atoms in 
        # their equilibrium distance in the crystal is zero it is only needed to calculate the energy of a single atom
        # in an empty system.
        
        # The crystal is created
        atoms = FaceCenteredCubic(size=(self.Size,self.Size,self.Size),symbol=self.Element)
        
        # a single atom is taken out of the crystal
        atoms2 = atoms[[0,]]

        # The calculator is attached to the atoms objects
        atoms.set_calculator(EMT)
        atoms2.set_calculator(EMT)

        # The energy difference between the atom alone in vacuum and in the crystal structure is calculated and returned
        return atoms2.get_potential_energy() - atoms.get_potential_energy() / len(atoms)
        

    def SurfaceEnergy_Calculator(self,EMT,PARAMETERS):
        """ The Method calculates and returns the surface energy for the given element along the [0,0,1] and [1,1,1] 
            directions in the FCC crystal structure. """
        # The size of the crystals are set " NEED TO THINK ABOUT THESE LATER! " 
        S001 = 3,3,5
        S111 = 5,5,5
        # The surfaces (slabs) are created (pbc=(1,1,0) creates periodic boudry conditions 
        # in two of three directions and thus leaves the last direction as two surfaces.
        Surface001 = FaceCenteredCubic(size=S001,symbol=self.Element,pbc=(1,1,0))
        Surface111 = FaceCenteredCubic(size=S111,directions=[[1,-1,0],[1,1,-2],[1,1,1]],
                                       symbol=self.Element,pbc=(1,1,0))
        Surface001.set_calculator(EMT)
        Surface111.set_calculator(EMT)

        # A structural relaxsation is run for the surface crystal in order to secure 
        # the correct structure of the crystal.
        dyn001 = BFGS(Surface001, trajectory='relaxedsurface001.traj')
        dyn111 = BFGS(Surface111, trajectory='relaxedsurface111.traj')
        dyn001.run(fmax=0.01)
        dyn111.run(fmax=0.01)        

        # The referance bulk crystals are created
        Bulk001 = FaceCenteredCubic(size=S001,symbol=self.Element)
        Bulk111 = FaceCenteredCubic(size=S111,directions=[[1,-1,0],[1,1,-2],[1,1,1]],symbol=self.Element)

        # The calculator is assigned
        Bulk001.set_calculator(EMT)
        Bulk111.set_calculator(EMT)

        # The surface area is calculated
        # The cross product between the x and y axis in the crystal is determined
        Cross001 =  numpy.cross(Bulk001.get_cell()[:,0],Bulk001.get_cell()[:,1])
        Cross111 =  numpy.cross(Bulk111.get_cell()[:,0],Bulk111.get_cell()[:,1])
        # The area of the surface is determined from the formular A = |X x Y|. 
        area001 = numpy.sqrt(numpy.dot(Cross001,Cross001))
        area111 = numpy.sqrt(numpy.dot(Cross111,Cross111))

        # The surface energy is calculated and returned (two surfaces are present in 
        # SurfaceRelaxed)
        return ( (Surface001.get_potential_energy() - Bulk001.get_potential_energy()) / 2 / area001,
                 (Surface111.get_potential_energy() - Bulk111.get_potential_energy()) / 2 / area111)

    def Surface001(self,EMT,PARAMETERS):
        """ The Method calculates and returns the surface energy for the given element along the [0,0,1] direction in   
            the FCC crystal structure. """
        # The size of the crystals are set: 
        S001 = 3,3,5
         # The surfaces (slabs) are created (pbc=(1,1,0) creates periodic boudry conditions 
        # in two of three directions and thus leaves the last direction as two surfaces.
        Surface001 = FaceCenteredCubic(size=S001,symbol=self.Element,pbc=(1,1,0))
        Surface001.set_calculator(EMT)

        # A structural relaxsation is run for the surface crystal in order to secure 
        # the correct structure of the crystal.
        dyn001 = BFGS(Surface001, logfile=None)
        dyn001.run(fmax=0.01)

        # The referance bulk crystals are created
        Bulk001 = FaceCenteredCubic(size=S001,symbol=self.Element)

        # The calculator is assigned
        Bulk001.set_calculator(EMT)
  
        # The surface area is calculated
        # The cross product between the x and y axis in the crystal is determined
        Cross001 =  numpy.cross(Bulk001.get_cell()[:,0],Bulk001.get_cell()[:,1])
        # The area of the surface is determined from the formular A = |X x Y|. 
        area001 = numpy.sqrt(numpy.dot(Cross001,Cross001))
    
        # The surface energy is calculated and returned (two surfaces are present in 
        # SurfaceRelaxed)
        return ( (Surface001.get_potential_energy() - Bulk001.get_potential_energy()) / 2 / area001)
   
    def Surface110(self,EMT,PARAMETERS):
        """ The Method calculates and returns the surface energy for the given element along the [1,1,0] direction in   
            the FCC crystal structure. """
        # The size of the crystals are set: 
        S110 = 3,5,5
         # The surfaces (slabs) are created (pbc=(1,1,0) creates periodic boudry conditions 
        # in two of three directions and thus leaves the last direction as two surfaces.
        Surface110 = FaceCenteredCubic(size=S110,directions=[[1,-1,0],[0,0,1],[1,1,0]],
                                       symbol=self.Element,pbc=(1,1,0))
        Surface110.set_calculator(EMT)

        # A structural relaxsation is run for the surface crystal in order to secure 
        # the correct structure of the crystal.
        dyn001 = BFGS(Surface110, logfile=None)
        dyn001.run(fmax=0.01)

        # The referance bulk crystals are created
        Bulk110 = FaceCenteredCubic(size=S110,directions=[[1,-1,0],[0,0,1],[1,1,0]],
                                    symbol=self.Element)

        # The calculator is assigned
        Bulk110.set_calculator(EMT)
  
        # The surface area is calculated
        # The cross product between the x and y axis in the crystal is determined
        Cross110 =  numpy.cross(Bulk110.get_cell()[:,0],Bulk110.get_cell()[:,1])
        # The area of the surface is determined from the formular A = |X x Y|. 
        area110 = numpy.sqrt(numpy.dot(Cross110,Cross110))
    
        # The surface energy is calculated and returned (two surfaces are present in 
        # SurfaceRelaxed)
        return ( (Surface110.get_potential_energy() - Bulk110.get_potential_energy()) / 2 / area110)
  
    def Surface111(self,EMT,PARAMETERS):
        """ The Method calculates and returns the surface energy for the given element along the [1,1,1] 
            direction in the FCC crystal structure. """
        # The size of the crystals are set " NEED TO THINK ABOUT THESE LATER! " 
        S111 = 4,4,4
        # The surfaces (slabs) are created (pbc=(1,1,0) creates periodic boudry conditions 
        # in two of three directions and thus leaves the last direction as two surfaces.
        Surface111 = FaceCenteredCubic(size=S111,directions=[[1,-1,0],[1,1,-2],[1,1,1]],
                                       symbol=self.Element,pbc=(1,1,0))
        Surface111.set_calculator(EMT)

        # A structural relaxsation is run for the surface crystal in order to secure 
        # the correct structure of the crystal.
        dyn111 = BFGS(Surface111, logfile=None)
        dyn111.run(fmax=0.01)        

        # The referance bulk crystals are created
        Bulk111 = FaceCenteredCubic(size=S111,directions=[[1,-1,0],[1,1,-2],[1,1,1]],symbol=self.Element)

        # The calculator is assigned
        Bulk111.set_calculator(EMT)

        # The surface area is calculated
        # The cross product between the x and y axis in the crystal is determined
        Cross111 =  numpy.cross(Bulk111.get_cell()[:,0],Bulk111.get_cell()[:,1])
        # The area of the surface is determined from the formular A = |X x Y|. 
        area111 = numpy.sqrt(numpy.dot(Cross111,Cross111))

        # The surface energy is calculated and returned (two surfaces are present in 
        # SurfaceRelaxed)
        return ( (Surface111.get_potential_energy() - Bulk111.get_potential_energy()) / 2 / area111)


    def L12_BM_LC_Calculator(self,EMT,PARAMETERS):
        # Return 0 is used when the method is not desired to be used
        # return 0
        """ The method L12_BM_LC_Calculator uses the EMT calculator to find the lattice constant which gives 
            the lowest possible energy for an alloy of two elements and from a polynomial fit of the energy as a 
            function of the lattice constant it calculates and returns the Bulk modulus, the volume of the system at 
            the lowest possible energy and this energy """
        # Three values for the lattice constant, a0, are chosen. The highest and lowest value is chosen so that it     
        # is certain that these are to high/low respectively compared to the "correct" value. This is done by using the 
        # experimental value of the lattice constant, a0_exp, as the middle value and to high/low values are then: 
        # a0_exp +- a0_mod, a0_mod = a0_exp/10
        a0_exp = ( ( PARAMETERS[self.Element[0]][1] + 3 * PARAMETERS[self.Element[1]][1] ) 
                     * numpy.sqrt(2) * beta * Bohr / 4 )
        a0_mod = a0_exp * 0.20 
        a0_guesses = numpy.array([a0_exp - a0_mod,a0_exp,a0_exp + a0_mod])
    
        # An atoms object consisting of atoms of the chosen element is initialized 
        atoms = L1_2(size=(self.Size,self.Size,self.Size),symbol=self.Element,latticeconstant=a0_exp)
        atoms.set_calculator(EMT)

        # An identity matrix for the system is saved to a variable
        IdentityMatrix = numpy.array([[1,0,0],[0,1,0],[0,0,1]])

        # An array for the energy for the chosen guesses for a0 is initialized:
        E_guesses = numpy.zeros(5)

        # The energies are calculated
        for i in range(3):
        # Changes the lattice constant for the atoms object
            atoms.set_cell(a0_guesses[i] * self.Size * IdentityMatrix,scale_atoms=True)
            # Calculates the energy of the system for the new lattice constant
            E_guesses[i] = atoms.get_potential_energy()
           
        # Bisection is used in order to find a small interval of the lattice constant for the minimum of the energy (an 
        # abitrary interval length is chosen), This is possible because we are certian from theory that there is only 
        # one minimum of the energy function in the interval of interest and thus we wont "fall" into a local minimum 
        # and stay there by accident.
        while (a0_guesses[2]-a0_guesses[0]) >= self.Tol:
            if min([E_guesses[0],E_guesses[2]]) == E_guesses[0]:
                # A new guess for the lattice constant is introduced
                a0_new_guess = 0.67 * a0_guesses[1] + 0.33 * a0_guesses[2]
                # The energy for this new guess is calculated
                atoms.set_cell(a0_new_guess * self.Size * IdentityMatrix,scale_atoms=True)
                E_new_guess = atoms.get_potential_energy()
                # A check for changes in the energy minimum is made and the guesses for a0 and their corrosponding
                # energies are ajusted.
                if min(E_new_guess,min(E_guesses[0:3])) != E_new_guess:
                    a0_guesses[2] = a0_new_guess
                    E_guesses[2] = E_new_guess 
                else:
                    a0_guesses[0] = a0_guesses[1]
                    a0_guesses[1] = a0_new_guess
                    E_guesses[0] = E_guesses[1]
                    E_guesses[1] = E_new_guess
         

            elif min([E_guesses[0],E_guesses[2]]) == E_guesses[2]: 
                # A new guess for the lattice constant is introduced
                a0_new_guess = 0.33 * a0_guesses[0] + 0.67 * a0_guesses[1]
                # The energy for this new guess is calculated
                atoms.set_cell(a0_new_guess * self.Size * IdentityMatrix,scale_atoms=True)
                E_new_guess = atoms.get_potential_energy()
                # A check for changes in the energy minimum is made and the guesses for a0 and their corrosponding
                # energies are ajusted.
                if min(E_new_guess,min(E_guesses[0:3])) != E_new_guess:
                    a0_guesses[0] = a0_new_guess
                    E_guesses[0] = E_new_guess 
                else:
                    a0_guesses[2] = a0_guesses[1]
                    a0_guesses[1] = a0_new_guess
                    E_guesses[2] = E_guesses[1]
                    E_guesses[1] = E_new_guess

        
        # An estimate of the minimum energy can now be found from a second degree polynomial fit through the three 
        # current guesses for a0 and the corresponding values of the energy.
        Poly = numpyfit.polyfit(a0_guesses,E_guesses[0:3],2)

        # The lattice constant corresponding to the lowest energy from the Polynomiel fit is found
        a0 = - Poly[1] / (2 * Poly[0])

        # Now five guesses for a0 and the corresponding energy are evaluated from and around the current a0.
        a0_guesses = a0 * numpy.array([1 - 2 * self.Tol / 5,1 - self.Tol / 5,1,1 + self.Tol / 5,1 + 2 * self.Tol / 5])


        for i in range(5):
            # Changes the lattice constant for the atoms object
            atoms.set_cell(a0_guesses[i] * self.Size * IdentityMatrix,scale_atoms=True)
            # Calculates the energy of the system for the new lattice constant
            E_guesses[i] = atoms.get_potential_energy()

        # The method EquationOfState is now used to find the Bulk modulus and the minimum energy for the system.

        # The volume of the sample for the given a0_guesses
        Vol = (self.Size * a0_guesses )**3
        
        # The equilibrium volume, energy and bulk modulus are calculated
        (Vol0 , E0 , B ) = EquationOfState(Vol.tolist(),E_guesses.tolist()).fit()



        return (Vol0,E0,B,Vol0**(1./3) / self.Size)

    def L12_BulkModulus(self,EMT,PARAMETERS):
        return self.L12_BM_LC_Calculator(EMT,PARAMETERS)[2]

    def L12_LatticeConstant(self,EMT,PARAMETERS):
        return self.L12_BM_LC_Calculator(EMT,PARAMETERS)[3]

  







"""Test the FixAtoms constraint and the Subset filter."""

from asap3 import *
from ase.lattice.cubic import FaceCenteredCubic
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3.md.verlet import VelocityVerlet
from asap3.md.nvtberendsen import NVTBerendsen
from asap3.constraints import FixAtoms, Filter
from ase.constraints import FixAtoms as ASE_FixAtoms
from asap3.test.pytest_markers import ReportTest, serial
from ase.parallel import world

import numpy as np

import pytest

def sanity(name, initial, final, fixed):
    print(f"Sanity check, {name}:")
    ok = (final == initial) + np.logical_not(fixed)[:,np.newaxis]
    assert ok.all(), 'Stationary atoms have not moved'
    ok = (final != initial) + fixed[:,np.newaxis]
    assert ok.all(), 'Mobile atoms have moved'

def sanity_parallel(name, atoms):
    initial = atoms.arrays['r_init']
    final = atoms.get_positions()
    fixed = atoms.get_tags().astype(bool)
    
    print(f"Sanity check, {name}:")
    ok = (final == initial) + np.logical_not(fixed)[:,np.newaxis]
    assert ok.all(), 'Stationary atoms have not moved'
    ok = (final != initial) + fixed[:,np.newaxis]
    assert ok.all(), 'Mobile atoms have moved'


@pytest.mark.parametrize('dynamics', ("Verlet", "NVTBerendsen", "Langevin"))
def test_change_constraint(dynamics, cpulayout):
    if cpulayout:
        # Parallel simulation
        if world.rank == 0:
            size = 5 * np.array(cpulayout)
            init = FaceCenteredCubic(size=size, symbol='Cu', pbc=False)
            z = init.get_positions()[:,2]
            fixedatoms = np.less(z, 0.501*z.max())
            print(len(init), sum(fixedatoms))
            MaxwellBoltzmannDistribution(init, temperature_K=6000)
            init.set_tags(fixedatoms)
        else:
            init = None
        nrun = 1000
    else:
        # Serial simulation
        init = FaceCenteredCubic(size=(10,10,10), symbol='Cu', pbc=False)
        z = init.get_positions()[:,2]
        fixedatoms = np.less(z, 0.501*z.max())
        print(len(init), sum(fixedatoms))
        init.set_tags(fixedatoms)
        MaxwellBoltzmannDistribution(init, temperature_K=2000)
        r_init = init.get_positions()
        nrun = 50

    print()
    print("Running simulation with Asap's FixAtoms")
    if cpulayout:
        atoms2 = MakeParallelAtoms(init, cpulayout)
        atoms2.arrays['r_init'] = atoms2.get_positions()
    else:
        atoms2 = Atoms(init)
    atoms2.calc = EMT()
    atoms2.set_constraint(FixAtoms(mask=atoms2.get_tags().astype(bool)))

    if dynamics == "Verlet":
        dyn = VelocityVerlet(atoms2, 3*units.fs)
    elif dynamics == "NVTBerendsen":
        dyn = NVTBerendsen(atoms2, 3*units.fs, temperature_K=2000, taut=200*units.fs)
    elif dynamics == "Langevin":
        dyn = Langevin(atoms2, 3*units.fs, temperature_K=2000, friction=1e-3)
    else:
        assert False
    dyn.run(50)
    r2 = atoms2.get_positions()

    if cpulayout:
        sanity_parallel(dynamics+"+FixAtoms", atoms2)
    else:
        sanity(dynamics+"+FixAtoms", r_init, r2, fixedatoms)

    x = r2[:,0]
    fixedatoms2 = np.less(x, 0.501 * x.max())
    atoms2.set_tags(fixedatoms2)
    atoms2.arrays['r_init'] = atoms2.get_positions()
    print(len(atoms2), sum(fixedatoms2))

    print("Running simulation with new FixAtoms")
    atoms2.set_constraint(FixAtoms(mask=fixedatoms2))
    #dyn = VelocityVerlet(atoms2, 2*units.fs)

    dyn.run(50)
    r3 = atoms2.get_positions()

    if cpulayout:
        sanity_parallel(dynamics+"+new-FixAtoms", atoms2)
    else:
        sanity(dynamics+"+new-FixAtoms", r2, r3, fixedatoms2)


if __name__ == '__main__':
    test_change_constraint('Verlet', [1, 1, world.size])




from ase.md.nose_hoover_chain import NoseHooverChainNVT as _NoseHooverChainNVT
from ase.md.nose_hoover_chain import NoseHooverChainThermostat as _NoseHooverChainThermostat
from ase.md.nose_hoover_chain import IsotropicMTKNPT as _IsotropicMTKNPT
try:
    from ase.md.nose_hoover_chain import MTKNPT as _MTKNPT
except ImportError:
    # Not yet in ASE
    class _MTKNPT:
        too_old_ase = True

import asap3.mpi
import ase
import numpy as np

class NoseHooverChainNVT(_NoseHooverChainNVT):
    def __init__(
        self,
        atoms: ase.Atoms,
        timestep: float,
        temperature_K: float,
        tdamp: float,
        tchain: int = 3,
        tloop: int = 1,
        communicator = asap3.mpi.world,
        **kwargs,
    ):
        # Replace parent's __init__, but chain to grandparent.
        super(_NoseHooverChainNVT, self).__init__(
            atoms=atoms, 
            timestep=timestep,
            **kwargs
        )
        self.comm = communicator
        if ase.__version__ < '3.25.0':
            raise RuntimeError('ASE too old.  NoseHooverChainNVT requires ASE 3.23.0.')
        
        num_atoms = self.atoms.get_global_number_of_atoms()
        self._thermostat = Asap_NoseHooverChainThermostat(
            num_atoms_global=num_atoms,
            communicator=communicator,
            masses=self.masses,
            temperature_K=temperature_K,
            tdamp=tdamp,
            tchain=tchain,
            tloop=tloop,
        )
        

        # The following variables are updated during self.step()
        # and synced with atoms before/after each force calculation.
        self._q = self.atoms.get_positions()
        self._p = self.atoms.get_momenta()

    def _get_forces(self) -> np.ndarray:
        f = super()._get_forces()
        # Sync internal variables as atoms may have migrated.
        self._p = self.atoms.get_momenta()
        self._q = self.atoms.get_positions()
        self.masses = self.atoms.get_masses()
        self.masses.shape = (-1, 1)
        self._thermostat.set_masses(self.masses)
        return f

class Asap_NoseHooverChainThermostat(_NoseHooverChainThermostat):
    def __init__(self, communicator, **kwargs):
        super().__init__(**kwargs)
        self.comm = communicator

    def set_masses(self, m):
        self._masses = m

    def _integrate_p_eta_j(self, p: np.ndarray, j: int, 
                           delta2: float, delta4: float) -> None:
        if j < self._tchain - 1:
            self._p_eta[j] *= np.exp(
                -delta4 * self._p_eta[j + 1] / self._Q[j + 1]
            )

        if j == 0:
            # Change compared to ASE
            p2overmass = np.sum(p**2 / self._masses)
            p2overmass = self.comm.sum(p2overmass)
            g_j = p2overmass - 3 * self._num_atoms_global * self._kT
        else:
            g_j = self._p_eta[j - 1] ** 2 / self._Q[j - 1] - self._kT
        self._p_eta[j] += delta2 * g_j

        if j < self._tchain - 1:
            self._p_eta[j] *= np.exp(
                -delta4 * self._p_eta[j + 1] / self._Q[j + 1]
            )

class IsotropicMTKNPT(_IsotropicMTKNPT):
    def __init__(
        self,
        atoms: ase.Atoms,
        timestep: float,
        temperature_K: float,
        pressure_au: float,
        tdamp: float,
        pdamp: float,
        tchain: int = 3,
        pchain: int = 3,
        tloop: int = 1,
        ploop: int = 1,
        communicator = asap3.mpi.world,
        **kwargs,
    ):
        """
        Parameters
        ----------
        atoms: ase.Atoms
            The atoms object.
        timestep: float
            The time step in ASE time units.
        temperature_K: float
            The target temperature in K.
        pressure_au: float
            The external pressure in eV/Å^3.
        tdamp: float
            The characteristic time scale for the thermostat in ASE time units.
            Typically, it is set to 100 times of `timestep`.
        pdamp: float
            The characteristic time scale for the barostat in ASE time units.
            Typically, it is set to 1000 times of `timestep`.
        tchain: int
            The number of thermostat variables in the Nose-Hoover thermostat.
        pchain: int
            The number of barostat variables in the MTK barostat.
        tloop: int
            The number of sub-steps in thermostat integration.
        ploop: int
            The number of sub-steps in barostat integration.
        **kwargs : dict, optional
            Additional arguments passed to :class:~ase.md.md.MolecularDynamics
            base class.
        """
        super().__init__(
            atoms=atoms,
            timestep=timestep,
            temperature_K=temperature_K,
            pressure_au=pressure_au,
            tdamp=tdamp,
            pdamp=pdamp,
            tchain=tchain,
            pchain=pchain,
            tloop=tloop,
            ploop=ploop,
            **kwargs,
        )
        self.comm = communicator

        # Replace the thermostat with Asap version.  The barostat is OK.
        num_atoms_global = self.atoms.get_global_number_of_atoms()
        self._thermostat = Asap_NoseHooverChainThermostat(
            communicator=communicator,
            num_atoms_global=num_atoms_global,
            masses=self.masses,
            temperature_K=temperature_K,
            tdamp=tdamp,
            tchain=tchain,
            tloop=tloop,
        )

    def _get_forces(self) -> np.ndarray:
        f = super()._get_forces()
        self._update_internal()
        return f

    def _update_internal(self):
        "Sync internal variables as atoms may have migrated."
        self._p = self.atoms.get_momenta()
        self._q = self.atoms.get_positions()
        self.masses = self.atoms.get_masses()
        self.masses.shape = (-1, 1)
        self._thermostat.set_masses(self.masses)

    def _get_pressure(self) -> np.ndarray:
        pressure = super()._get_pressure()
        self._update_internal()
        return pressure

    def _integrate_p_cell(self, delta: float) -> None:
        """Integrate exp(i * L_(epsilon, 2) * delta)"""
        pressure = self._get_pressure()
        volume = self._get_volume()
        p2overmass = np.sum(self._p**2 / self.masses)
        p2overmass = self.comm.sum(p2overmass)
        G = (
            3 * volume * (pressure - self._pressure_au)
            + p2overmass / self.atoms.get_global_number_of_atoms()
        )
        self._p_eps += delta * G

class MTKNPT(_MTKNPT):
    """Isothermal-isobaric molecular dynamics with volume-and-cell fluctuations
    by Martyna-Tobias-Klein (MTK) method [1].

    See also `NoseHooverChainNVT` for the references.

    - [1] G. J. Martyna, D. J. Tobias, and M. L. Klein, J. Chem. Phys. 101,
          4177-4189 (1994). https://doi.org/10.1063/1.467468
    """
    def __init__(
        self,
        atoms: ase.Atoms,
        timestep: float,
        temperature_K: float,
        pressure_au: float,
        tdamp: float,
        pdamp: float,
        tchain: int = 3,
        pchain: int = 3,
        tloop: int = 1,
        ploop: int = 1,
        communicator = asap3.mpi.world,
        **kwargs,
    ):
        """
        Parameters
        ----------
        atoms: ase.Atoms
            The atoms object.
        timestep: float
            The time step in ASE time units.
        temperature_K: float
            The target temperature in K.
        pressure_au: float
            The external pressure in eV/Ang^3.
        tdamp: float
            The characteristic time scale for the thermostat in ASE time units.
            Typically, it is set to 100 times of `timestep`.
        pdamp: float
            The characteristic time scale for the barostat in ASE time units.
            Typically, it is set to 1000 times of `timestep`.
        tchain: int
            The number of thermostat variables in the Nose-Hoover thermostat.
        pchain: int
            The number of barostat variables in the MTK barostat.
        tloop: int
            The number of sub-steps in thermostat integration.
        ploop: int
            The number of sub-steps in barostat integration.
        **kwargs : dict, optional
            Additional arguments passed to :class:~ase.md.md.MolecularDynamics
            base class.
        """
        super().__init__(atoms, 
                         timestep, 
                         temperature_K, 
                         pressure_au, 
                         tdamp, 
                         pdamp, 
                         tchain, 
                         pchain, 
                         tloop, 
                         ploop, 
                         *kwargs
        )
        self.comm = communicator

        # Replace the thermostat with Asap version.  The barostat is OK.
        num_atoms_global = self.atoms.get_global_number_of_atoms()
        self._thermostat = Asap_NoseHooverChainThermostat(
            communicator=communicator,
            num_atoms_global=num_atoms_global,
            masses=self.masses,
            temperature_K=temperature_K,
            tdamp=tdamp,
            tchain=tchain,
            tloop=tloop,
        )

    def _get_forces(self) -> np.ndarray:
        f = super()._get_forces()
        self._update_internal()
        return f

    def _update_internal(self):
        "Sync internal variables as atoms may have migrated."
        self._p = self.atoms.get_momenta()
        self._q = self.atoms.get_positions()
        self.masses = self.atoms.get_masses()
        self.masses.shape = (-1, 1)
        self._thermostat.set_masses(self.masses)

    def _get_stress(self) -> np.ndarray:
        stress = super()._get_stress()
        self._update_internal()
        return stress

    def _integrate_p_cell(self, delta: float) -> None:
        """Integrate exp(i * L_(g, 2) * delta)"""
        stress = self._get_stress()
        p2overmass = np.sum(self._p**2 / self.masses)
        p2overmass = self.comm.sum(p2overmass)
        G = (
            self._get_volume() * (stress - self._pressure_au * np.eye(3))
            + p2overmass / (3 * self.atoms.get_global_number_of_atoms())
                * np.eye(3)
        )
        self._p_g += delta * G

# Remove half-baked class if not supported by ASE
if hasattr(MTKNPT, 'too_old_ase'):
    del MTKNPT



from asap3 import *
from asap3.md.verlet import VelocityVerlet
from ase.lattice.cubic import FaceCenteredCubic
from asap3.io.trajectory import *
from ase.parallel import world
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution, Stationary
from asap3.constraints import FixAtoms
from asap3.test.pytest_markers import ReportTest
import numpy as np

import pytest


def test_Fixatoms_temperature(cpulayout):
    ismaster = world.rank == 0
    isparallel = world.size != 1

    temp = 3000  # K

    if ismaster:
        initial = FaceCenteredCubic(size=(10,10,10), symbol="Cu", pbc=(1,0,0))
    else:
        initial = None
    if isparallel:
        atoms = MakeParallelAtoms(initial, cpulayout)
        atoms.set_constraint(FixAtoms(mask=(atoms.get_ids() < atoms.get_global_number_of_atoms()/2)))
    else:
        atoms = initial.copy()
        atoms.set_constraint(FixAtoms(indices=range(len(atoms)//2)))

    atoms.calc = EMT()
    MaxwellBoltzmannDistribution(atoms, temperature_K=temp)
    Stationary(atoms)

    dyn = Langevin(atoms, 3*units.fs, temperature_K=temp, friction=0.02)
    dyn.run(1000)

    temps = [atoms.get_temperature()]
    for i in range(10):
        dyn.run(10)
        temps.append(atoms.get_temperature())
        
    avgtemp = np.mean(temps)
    print(f"Average temperature T={avgtemp:.1f}K  (task {world.rank})")
    c = atoms.constraints[0]
    print(f"Removed DOF: {c.get_removed_dof(atoms)}   (task {world.rank})")
    ReportTest("Mean temperature:", avgtemp, temp, 200)



import pickle
import sys

import numpy as np
import pytest

from asap3 import Atoms, VelocityVerlet, EMT, MDLogger, OpenKIMsupported, AsapError
if OpenKIMsupported:
    from asap3 import OpenKIMcalculator
from asap3.test.pytest_markers import serial, withmpi, withOpenKIM
from ase.parallel import world

@pytest.fixture
def data(datadir):
    picklehack = {'encoding': 'latin1'}
    picklefile = datadir / "Verlet.pickle"
    return pickle.load(open(picklefile, "rb"), **picklehack)

@pytest.mark.core
@serial
def test_verlet(data):
    runtest_verlet(data, EMT())

def runtest_verlet(data, potential):
    timeunit = 1.018047e-14             # Seconds
    femtosecond = 1e-15 / timeunit      # Marginally different from units.fs
    stresshack = {'include_ideal_gas': True}

    init_pos = np.array(data["initial"])
    init_pos.shape = (-1,3)
    init_box = np.array(data["box"])
    init_box.shape = (3,3)
    atoms = Atoms(positions=init_pos, cell=init_box)
    atoms.set_atomic_numbers(47 * np.ones((len(atoms),)))
    atoms.calc = potential

    dyn = VelocityVerlet(atoms, 2 * femtosecond)
    dyn.attach(MDLogger(dyn, atoms, '-', peratom=True), interval=5)

    etot = None

    for i in range(10):
        dyn.run(20)
        epot = atoms.get_potential_energy() / len(atoms)
        ekin = atoms.get_kinetic_energy() / len(atoms)
        if etot is None:
            etot = epot + ekin
        else:
            assert np.isclose(epot + ekin, etot, atol=1e-3), 'Energy conservation'
            
    final_pos = np.array(data["final"])
    diff = max(abs(atoms.get_positions().flat - final_pos))
    print("Maximal deviation of positions:", diff)
    assert diff < 1e-9, 'Maximal deviation of positions'

    diff = max(abs(atoms.get_stresses(**stresshack).flat - np.array(data["stress"])))

    print("Maximal deviation of stresses:", diff)
    assert diff < 1e-9, 'Maximal deviation of stresses'


@withmpi
def test_verlet_par(data, cpulayout):
    runtest_verlet_par(data, cpulayout, EMT())

@withmpi
@withOpenKIM
def test_verlet_KIM(data, cpulayout, openkimmodel, in_tmp_dir):
    try:
        calc = OpenKIMcalculator(openkimmodel)
    except AsapError as oops:
        if oops.args[0].startswith('Failed to initialize OpenKIM model'):
            calc = None
        else:
            raise
    if calc is not None:
        runtest_verlet_par(data, cpulayout, calc, epotexpected=-3.456898)

def runtest_verlet_par(data, cpulayout, potential, epotexpected=0.058370, report=False):
    from asap3 import MakeParallelAtoms

    # Ensure same time step as in Asap-2
    timeunit = 1.018047e-14             # Seconds
    femtosecond = 1e-15 / timeunit      # Marginally different from units.fs
    isparallel = world.size > 1
    ismaster = world.rank == 0

    if isparallel:
        print("RUNNING PARALLEL VERSION OF TEST SCRIPT")
    else:
        print("RUNNING SERIAL VERSION OF TEST SCRIPT")

    if ismaster:
        init_pos = np.array(data["initial"])
        init_pos.shape = (-1,3)
        init_box = np.array(data["box"])
        init_box.shape = (3,3)
        a = Atoms(positions=init_pos, cell=init_box, pbc=(1,1,1))
        a.set_atomic_numbers(29 * np.ones(len(a)))
        #atoms = a.repeat((2,2,3))
        # Ensure same order of atoms as in ASE-2
        atoms = a.repeat((1,1,3)).repeat((1,2,1)).repeat((2,1,1))
        dx = 0.1 * np.sin(np.arange(3*len(atoms))/10.0)
        dx.shape = (-1,3)
        atoms.set_positions(atoms.get_positions() + dx)
        del dx
        out = sys.stderr
    else:
        atoms = None
        out = open("/dev/null", "w")

    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)
        nTotalAtoms = atoms.get_global_number_of_atoms()
    else:
        nTotalAtoms = len(atoms)

    if report:
        report_verlet(out)

    print("Setting potential")
    atoms.calc = potential

    dyn = VelocityVerlet(atoms, 2 * femtosecond)
    # if isparallel:
    #     traj = ParallelNetCDFTrajectory("ptraj.nc", atoms, interval=20)
    # else:
    #     traj = NetCDFTrajectory("ptraj.nc", atoms, interval=20)
    # dyn.Attach(traj)
    # traj.Update()

    print("Number of atoms:", nTotalAtoms)

    epot = atoms.get_potential_energy() / nTotalAtoms
    ekin = atoms.get_kinetic_energy() / nTotalAtoms
    etotallist = [epot+ekin]
    ekinlist = [ekin]

    #report()

    if ismaster:
        print("\nE_pot = {:<12.5f}  E_kin = {:<12.5f}  E_tot = {:<12.5f}".format(epot, ekin,
                                                                    epot+ekin))
    assert np.isclose(epot, epotexpected, atol=1e-4), 'Initial potential energy'
    assert np.isclose(ekin, 0.0, atol=1e-9), 'Initial kinetic energy'

    dyn.attach(MDLogger(dyn, atoms, "-", peratom=True), interval=10)

    for i in range(40):
        dyn.run(5)
        epot = atoms.get_potential_energy() / nTotalAtoms
        ekin = atoms.get_kinetic_energy() / nTotalAtoms
        etotallist.append(epot+ekin)
        ekinlist.append(ekin)

    if ismaster:
        print("Average total energy:", sum(etotallist)/len(etotallist))
        print("Average kinetic energy:", sum(ekinlist)/len(ekinlist))

    assert np.isclose(sum(etotallist)/len(etotallist), epotexpected, atol=0.0001), 'Agv. total energy'
    assert np.isclose(sum(ekinlist)/len(ekinlist), 0.0290841, atol=0.002), 'Agv. kinetic energy' 

def report_verlet(out):
    "Debugging function - not normally used."
    for i in range(world.size):
        if i == world.rank:
            print("Data on processor", i, file=out)
            for key in atoms.arrays.keys():
                print("  ", key, atoms.arrays[key].shape, file=out)
            r = atoms.get_positions()
            if len(r):
                print("Limits to the positions:", file=out)
                print ("[%.4f, %.4f]  [%.4f, %.4f]  [%.4f, %.4f]" %
                       (min(r[:,0]), max(r[:,0]), min(r[:,1]), max(r[:,1]),
                        min(r[:,2]), max(r[:,2])), file=out)
            if world.size > 1:
                print("Ghost data on processor", i)
                for key in atoms.ghosts.keys():
                    print("  ", key, atoms.ghosts[key].shape, file=out)
                r = atoms.ghosts['positions']
                if len(r):
                    print("Limits to the ghost positions:", file=out)
                    print ("[%.4f, %.4f]  [%.4f, %.4f]  [%.4f, %.4f]" %
                           (min(r[:,0]), max(r[:,0]), min(r[:,1]), max(r[:,1]),
                            min(r[:,2]), max(r[:,2])), file=out)
        world.barrier()


